import logging


class AsyncioErrors(AssertionError):
    def __repr__(self):
        #  pylint: disable=unsubscriptable-object
        return "<AsyncioErrors: Got asyncio errors: %r" % self.args[0]


class Handler(logging.Handler):
    def __init__(self):
        super().__init__(level=logging.ERROR)
        self.messages = []

    def emit(self, record):
        message = record.msg % record.args
        self.messages.append(message)


asyncio_logger = logging.getLogger('asyncio')
handler = Handler()
asyncio_logger.addHandler(handler)



"""
    Tests for message properties for basic deliver
"""

import asyncio
import asynctest
import logging

from . import testcase


logger = logging.getLogger(__name__)


class ReplyTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def _server(self, server_future, exchange_name, routing_key):
        """Consume messages and reply to them by publishing messages back to the client using
        routing key set to the reply_to property"""
        server_queue_name = 'server_queue'
        await self.channel.queue_declare(server_queue_name, exclusive=False, no_wait=False)
        await self.channel.exchange_declare(exchange_name, type_name='direct')
        await self.channel.queue_bind(
            server_queue_name, exchange_name, routing_key=routing_key)

        async def server_callback(channel, body, envelope, properties):
            logger.debug('Server received message')
            server_future.set_result((body, envelope, properties))
            publish_properties = {'correlation_id': properties.correlation_id}
            logger.debug('Replying to %r', properties.reply_to)
            await self.channel.publish(
                b'reply message', exchange_name, properties.reply_to, publish_properties)
            logger.debug('Server replied')
        await self.channel.basic_consume(server_callback, queue_name=server_queue_name)
        logger.debug('Server consuming messages')


    async def _client(
            self, client_future, exchange_name, server_routing_key, correlation_id,
            client_routing_key):
        """Declare a queue, bind client_routing_key to it and publish a message to the server with
        the reply_to property set to that routing key"""
        client_queue_name = 'client_reply_queue'
        client_channel = await self.create_channel()
        await client_channel.queue_declare(
            client_queue_name, exclusive=True, no_wait=False)
        await client_channel.queue_bind(
            client_queue_name, exchange_name, routing_key=client_routing_key)

        async def client_callback(channel, body, envelope, properties):
            logger.debug('Client received message')
            client_future.set_result((body, envelope, properties))
        await client_channel.basic_consume(client_callback, queue_name=client_queue_name)
        logger.debug('Client consuming messages')

        await client_channel.publish(
            b'client message',
            exchange_name,
            server_routing_key,
            {'correlation_id': correlation_id, 'reply_to': client_routing_key})
        logger.debug('Client published message')

    async def test_reply_to(self):
        exchange_name = 'exchange_name'
        server_routing_key = 'reply_test'

        server_future = asyncio.Future(loop=self.loop)
        await self._server(server_future, exchange_name, server_routing_key)

        correlation_id = 'secret correlation id'
        client_routing_key = 'secret_client_key'

        client_future = asyncio.Future(loop=self.loop)
        await self._client(
            client_future, exchange_name, server_routing_key, correlation_id, client_routing_key)

        logger.debug('Waiting for server to receive message')
        server_body, server_envelope, server_properties = await server_future
        self.assertEqual(server_body, b'client message')
        self.assertEqual(server_properties.correlation_id, correlation_id)
        self.assertEqual(server_properties.reply_to, client_routing_key)
        self.assertEqual(server_envelope.routing_key, server_routing_key)

        logger.debug('Waiting for client to receive message')
        client_body, client_envelope, client_properties = await client_future
        self.assertEqual(client_body, b'reply message')
        self.assertEqual(client_properties.correlation_id, correlation_id)
        self.assertEqual(client_envelope.routing_key, client_routing_key)



import asynctest

from aioamqp.protocol import OPEN, CLOSED
from aioamqp.exceptions import AmqpClosedConnection

from . import testcase


class CloseTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_close(self):
        amqp = self.amqp
        self.assertEqual(amqp.state, OPEN)
        # grab a ref here because py36 sets _stream_reader to None in
        # StreamReaderProtocol.connection_lost()
        transport = amqp._stream_reader._transport
        await amqp.close()
        self.assertEqual(amqp.state, CLOSED)
        if hasattr(transport, 'is_closing'):
            self.assertTrue(transport.is_closing())
        else:
            # TODO: remove with python <3.4.4 support
            self.assertTrue(transport._closing)
        # make sure those 2 tasks/futures are properly set as finished
        await amqp.stop_now
        await amqp.worker

    async def test_multiple_close(self):
        amqp = self.amqp
        await amqp.close()
        self.assertEqual(amqp.state, CLOSED)
        with self.assertRaises(AmqpClosedConnection):
            await amqp.close()



"""
    Tests the "Channel" amqp class implementation
"""

import os
import unittest

import asynctest

from . import testcase
from .. import exceptions

IMPLEMENT_CHANNEL_FLOW = os.environ.get('IMPLEMENT_CHANNEL_FLOW', False)

class ChannelTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    _multiprocess_can_split_ = True

    async def test_open(self):
        channel = await self.amqp.channel()
        self.assertNotEqual(channel.channel_id, 0)
        self.assertTrue(channel.is_open)

    async def test_close(self):
        channel = await self.amqp.channel()
        result = await channel.close()
        self.assertEqual(result, True)
        self.assertFalse(channel.is_open)

    async def test_server_initiated_close(self):
        channel = await self.amqp.channel()
        try:
            await channel.basic_get(queue_name='non-existant')
        except exceptions.ChannelClosed as e:
            self.assertEqual(e.code, 404)
        self.assertFalse(channel.is_open)
        channel = await self.amqp.channel()

    async def test_alreadyclosed_channel(self):
        channel = await self.amqp.channel()
        result = await channel.close()
        self.assertEqual(result, True)

        with self.assertRaises(exceptions.ChannelClosed):
            result = await channel.close()

    async def test_multiple_open(self):
        channel1 = await self.amqp.channel()
        channel2 = await self.amqp.channel()
        self.assertNotEqual(channel1.channel_id, channel2.channel_id)

    async def test_channel_active_flow(self):
        channel = await self.amqp.channel()
        result = await channel.flow(active=True)
        self.assertTrue(result['active'])

    @unittest.skipIf(IMPLEMENT_CHANNEL_FLOW is False, "active=false is not implemented in RabbitMQ")
    async def test_channel_inactive_flow(self):
        channel = await self.amqp.channel()
        result = await channel.flow(active=False)
        self.assertFalse(result['active'])
        result = await channel.flow(active=True)

    async def test_channel_active_flow_twice(self):
        channel = await self.amqp.channel()
        result = await channel.flow(active=True)
        self.assertTrue(result['active'])
        result = await channel.flow(active=True)

    @unittest.skipIf(IMPLEMENT_CHANNEL_FLOW is False, "active=false is not implemented in RabbitMQ")
    async def test_channel_active_inactive_flow(self):
        channel = await self.amqp.channel()
        result = await channel.flow(active=True)
        self.assertTrue(result['active'])
        result = await channel.flow(active=False)
        self.assertFalse(result['active'])


class ChannelIdTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_channel_id_release_close(self):
        channels_count_start = self.amqp.channels_ids_count
        channel = await self.amqp.channel()
        self.assertEqual(self.amqp.channels_ids_count, channels_count_start + 1)
        result = await channel.close()
        self.assertEqual(result, True)
        self.assertFalse(channel.is_open)
        self.assertEqual(self.amqp.channels_ids_count, channels_count_start)



"""
    Amqp Protocol
"""

import asyncio
import logging

import pamqp.frame
import pamqp.heartbeat
import pamqp.specification

from . import channel as amqp_channel
from . import constants as amqp_constants
from . import frame as amqp_frame
from . import exceptions
from . import version


logger = logging.getLogger(__name__)


CONNECTING, OPEN, CLOSING, CLOSED = range(4)


class _StreamWriter(asyncio.StreamWriter):

    def write(self, data):
        super().write(data)
        self._protocol._heartbeat_timer_send_reset()

    def writelines(self, data):
        super().writelines(data)
        self._protocol._heartbeat_timer_send_reset()

    def write_eof(self):
        ret = super().write_eof()
        self._protocol._heartbeat_timer_send_reset()
        return ret


class AmqpProtocol(asyncio.StreamReaderProtocol):
    """The AMQP protocol for asyncio.

    See http://docs.python.org/3.4/library/asyncio-protocol.html#protocols for more information
    on asyncio's protocol API.

    """

    CHANNEL_FACTORY = amqp_channel.Channel

    def __init__(self, *args, **kwargs):
        """Defines our new protocol instance

        Args:
            channel_max: int, specifies highest channel number that the server permits.
                              Usable channel numbers are in the range 1..channel-max.
                              Zero indicates no specified limit.
            frame_max: int, the largest frame size that the server proposes for the connection,
                            including frame header and end-byte. The client can negotiate a lower value.
                            Zero means that the server does not impose any specific limit
                            but may reject very large frames if it cannot allocate resources for them.
            heartbeat: int, the delay, in seconds, of the connection heartbeat that the server wants.
                            Zero means the server does not want a heartbeat.
            loop: Asyncio.Eventloop: specify the eventloop to use.
            client_properties: dict, client-props to tune the client identification
        """
        self._loop = kwargs.get('loop') or asyncio.get_event_loop()
        self._reader = asyncio.StreamReader(loop=self._loop)
        super().__init__(self._reader, loop=self._loop)
        self._on_error_callback = kwargs.get('on_error')

        self.client_properties = kwargs.get('client_properties', {})
        self.connection_tunning = {}
        if 'channel_max' in kwargs:
            self.connection_tunning['channel_max'] = kwargs.get('channel_max')
        if 'frame_max' in kwargs:
            self.connection_tunning['frame_max'] = kwargs.get('frame_max')
        if 'heartbeat' in kwargs:
            self.connection_tunning['heartbeat'] = kwargs.get('heartbeat')

        self.connecting = asyncio.Future(loop=self._loop)
        self.connection_closed = asyncio.Event(loop=self._loop)
        self.stop_now = asyncio.Future(loop=self._loop)
        self.state = CONNECTING
        self.version_major = None
        self.version_minor = None
        self.server_properties = None
        self.server_mechanisms = None
        self.server_locales = None
        self.worker = None
        self.server_heartbeat = None
        self._heartbeat_timer_recv = None
        self._heartbeat_timer_send = None
        self._heartbeat_trigger_send = asyncio.Event(loop=self._loop)
        self._heartbeat_worker = None
        self.channels = {}
        self.server_frame_max = None
        self.server_channel_max = None
        self.channels_ids_ceil = 0
        self.channels_ids_free = set()
        self._drain_lock = asyncio.Lock(loop=self._loop)

    def connection_made(self, transport):
        super().connection_made(transport)
        self._stream_writer = _StreamWriter(transport, self, self._stream_reader, self._loop)

    def eof_received(self):
        super().eof_received()
        # Python 3.5+ started returning True here to keep the transport open.
        # We really couldn't care less so keep the behavior from 3.4 to make
        # sure connection_lost() is called.
        return False

    def connection_lost(self, exc):
        if exc is not None:
            logger.warning("Connection lost exc=%r", exc)
        self.connection_closed.set()
        self.state = CLOSED
        self._close_channels(exception=exc)
        self._heartbeat_stop()
        super().connection_lost(exc)

    def data_received(self, data):
        self._heartbeat_timer_recv_reset()
        super().data_received(data)

    async def ensure_open(self):
        # Raise a suitable exception if the connection isn't open.
        # Handle cases from the most common to the least common.

        if self.state == OPEN:
            return

        if self.state == CLOSED:
            raise exceptions.AmqpClosedConnection()

        # If the closing handshake is in progress, let it complete.
        if self.state == CLOSING:
            await self.wait_closed()
            raise exceptions.AmqpClosedConnection()

        # Control may only reach this point in buggy third-party subclasses.
        assert self.state == CONNECTING
        raise exceptions.AioamqpException("connection isn't established yet.")

    async def _drain(self):
        async with self._drain_lock:
            # drain() cannot be called concurrently by multiple coroutines:
            # http://bugs.python.org/issue29930. Remove this lock when no
            # version of Python where this bugs exists is supported anymore.
            await self._stream_writer.drain()

    async def _write_frame(self, channel_id, request, drain=True):
        amqp_frame.write(self._stream_writer, channel_id, request)
        if drain:
            await self._drain()

    async def close(self, no_wait=False, timeout=None):
        """Close connection (and all channels)"""
        await self.ensure_open()
        self.state = CLOSING
        request = pamqp.specification.Connection.Close(
            reply_code=0,
            reply_text='',
            class_id=0,
            method_id=0
        )

        await self._write_frame(0, request)
        if not no_wait:
            await self.wait_closed(timeout=timeout)

    async def wait_closed(self, timeout=None):
        await asyncio.wait_for(self.connection_closed.wait(), timeout=timeout, loop=self._loop)
        if self._heartbeat_worker is not None:
            try:
                await asyncio.wait_for(self._heartbeat_worker, timeout=timeout, loop=self._loop)
            except asyncio.CancelledError:
                pass

    async def close_ok(self, frame):
        self._stream_writer.close()

    async def start_connection(self, host, port, login, password, virtualhost, ssl=False,
            login_method='PLAIN', insist=False):
        """Initiate a connection at the protocol level
            We send `PROTOCOL_HEADER'
        """

        if login_method != 'PLAIN':
            logger.warning('only PLAIN login_method is supported, falling back to AMQPLAIN')

        self._stream_writer.write(amqp_constants.PROTOCOL_HEADER)

        # Wait 'start' method from the server
        await self.dispatch_frame()

        client_properties = {
            'capabilities': {
                'consumer_cancel_notify': True,
                'connection.blocked': False,
            },
            'copyright': 'BSD',
            'product': version.__package__,
            'product_version': version.__version__,
        }
        client_properties.update(self.client_properties)
        auth = {
            'LOGIN': login,
            'PASSWORD': password,
        }

        # waiting reply start with credentions and co
        await self.start_ok(client_properties, 'PLAIN', auth, self.server_locales)

        # wait for a "tune" reponse
        await self.dispatch_frame()

        tune_ok = {
            'channel_max': self.connection_tunning.get('channel_max', self.server_channel_max),
            'frame_max': self.connection_tunning.get('frame_max', self.server_frame_max),
            'heartbeat': self.connection_tunning.get('heartbeat', self.server_heartbeat),
        }
        # "tune" the connexion with max channel, max frame, heartbeat
        await self.tune_ok(**tune_ok)

        # update connection tunning values
        self.server_frame_max = tune_ok['frame_max']
        self.server_channel_max = tune_ok['channel_max']
        self.server_heartbeat = tune_ok['heartbeat']

        if self.server_heartbeat > 0:
            self._heartbeat_timer_recv_reset()
            self._heartbeat_timer_send_reset()

        # open a virtualhost
        await self.open(virtualhost, capabilities='', insist=insist)

        # wait for open-ok
        channel, frame = await self.get_frame()
        await self.dispatch_frame(channel, frame)

        await self.ensure_open()
        # for now, we read server's responses asynchronously
        self.worker = asyncio.ensure_future(self.run(), loop=self._loop)

    async def get_frame(self):
        """Read the frame, and only decode its header

        """
        return await amqp_frame.read(self._stream_reader)

    async def dispatch_frame(self, frame_channel=None, frame=None):
        """Dispatch the received frame to the corresponding handler"""

        method_dispatch = {
            pamqp.specification.Connection.Close.name: self.server_close,
            pamqp.specification.Connection.CloseOk.name: self.close_ok,
            pamqp.specification.Connection.Tune.name: self.tune,
            pamqp.specification.Connection.Start.name: self.start,
            pamqp.specification.Connection.OpenOk.name: self.open_ok,
        }
        if frame_channel is None and frame is None:
            frame_channel, frame = await self.get_frame()

        if isinstance(frame, pamqp.heartbeat.Heartbeat):
            return

        if frame_channel != 0:
            channel = self.channels.get(frame_channel)
            if channel is not None:
                await channel.dispatch_frame(frame)
            else:
                logger.info("Unknown channel %s", frame_channel)
            return

        if frame.name not in method_dispatch:
            logger.info("frame %s is not handled", frame.name)
            return
        await method_dispatch[frame.name](frame)

    def release_channel_id(self, channel_id):
        """Called from the channel instance, it relase a previously used
        channel_id
        """
        self.channels_ids_free.add(channel_id)

    @property
    def channels_ids_count(self):
        return self.channels_ids_ceil - len(self.channels_ids_free)

    def _close_channels(self, reply_code=None, reply_text=None, exception=None):
        """Cleanly close channels

            Args:
                reply_code:     int, the amqp error code
                reply_text:     str, the text associated to the error_code
                exc:            the exception responsible of this error

        """
        if exception is None:
            exception = exceptions.ChannelClosed(reply_code, reply_text)

        if self._on_error_callback:
            if asyncio.iscoroutinefunction(self._on_error_callback):
                asyncio.ensure_future(self._on_error_callback(exception), loop=self._loop)
            else:
                self._on_error_callback(exceptions.ChannelClosed(exception))

        for channel in self.channels.values():
            channel.connection_closed(reply_code, reply_text, exception)

    async def run(self):
        while not self.stop_now.done():
            try:
                await self.dispatch_frame()
            except exceptions.AmqpClosedConnection as exc:
                logger.info("Close connection")
                self.stop_now.set_result(None)

                self._close_channels(exception=exc)
            except Exception:  # pylint: disable=broad-except
                logger.exception('error on dispatch')

    async def heartbeat(self):
        """ deprecated heartbeat coroutine

        This coroutine is now a no-op as the heartbeat is handled directly by
        the rest of the AmqpProtocol class.  This is kept around for backwards
        compatibility purposes only.
        """
        await self.stop_now

    async def send_heartbeat(self):
        """Sends an heartbeat message.
        It can be an ack for the server or the client willing to check for the
        connexion timeout
        """
        request = pamqp.heartbeat.Heartbeat()
        await self._write_frame(0, request)

    def _heartbeat_timer_recv_timeout(self):
        # 4.2.7 If a peer detects no incoming traffic (i.e. received octets) for
        # two heartbeat intervals or longer, it should close the connection
        # without following the Connection.Close/Close-Ok handshaking, and log
        # an error.
        # TODO(rcardona) raise a "timeout" exception somewhere
        self._stream_writer.close()

    def _heartbeat_timer_recv_reset(self):
        if self.server_heartbeat is None:
            return
        if self._heartbeat_timer_recv is not None:
            self._heartbeat_timer_recv.cancel()
        self._heartbeat_timer_recv = self._loop.call_later(
            self.server_heartbeat * 2,
            self._heartbeat_timer_recv_timeout)

    def _heartbeat_timer_send_reset(self):
        if self.server_heartbeat is None:
            return
        if self._heartbeat_timer_send is not None:
            self._heartbeat_timer_send.cancel()
        self._heartbeat_timer_send = self._loop.call_later(
            self.server_heartbeat,
            self._heartbeat_trigger_send.set)
        if self._heartbeat_worker is None:
            self._heartbeat_worker = asyncio.ensure_future(self._heartbeat(), loop=self._loop)

    def _heartbeat_stop(self):
        self.server_heartbeat = None
        if self._heartbeat_timer_recv is not None:
            self._heartbeat_timer_recv.cancel()
        if self._heartbeat_timer_send is not None:
            self._heartbeat_timer_send.cancel()
        if self._heartbeat_worker is not None:
            self._heartbeat_worker.cancel()

    async def _heartbeat(self):
        while self.state != CLOSED:
            await self._heartbeat_trigger_send.wait()
            self._heartbeat_trigger_send.clear()
            await self.send_heartbeat()

    # Amqp specific methods
    async def start(self, frame):
        """Method sent from the server to begin a new connection"""
        self.version_major = frame.version_major
        self.version_minor = frame.version_minor
        self.server_properties = frame.server_properties
        self.server_mechanisms = frame.mechanisms
        self.server_locales = frame.locales

    async def start_ok(self, client_properties, mechanism, auth, locale):
        def credentials():
            return '\0{LOGIN}\0{PASSWORD}'.format(**auth)

        request = pamqp.specification.Connection.StartOk(
            client_properties=client_properties,
            mechanism=mechanism,
            locale=locale,
            response=credentials()
        )
        await self._write_frame(0, request)

    async def server_close(self, frame):
        """The server is closing the connection"""
        self.state = CLOSING
        reply_code = frame.reply_code
        reply_text = frame.reply_text
        class_id = frame.class_id
        method_id = frame.method_id
        logger.warning("Server closed connection: %s, code=%s, class_id=%s, method_id=%s",
            reply_text, reply_code, class_id, method_id)
        self._close_channels(reply_code, reply_text)
        await self._close_ok()
        self._stream_writer.close()

    async def _close_ok(self):
        request = pamqp.specification.Connection.CloseOk()
        await self._write_frame(0, request)

    async def tune(self, frame):
        self.server_channel_max = frame.channel_max
        self.server_frame_max = frame.frame_max
        self.server_heartbeat = frame.heartbeat

    async def tune_ok(self, channel_max, frame_max, heartbeat):
        request = pamqp.specification.Connection.TuneOk(
            channel_max, frame_max, heartbeat
        )
        await self._write_frame(0, request)

    async def secure_ok(self, login_response):
        pass

    async def open(self, virtual_host, capabilities='', insist=False):
        """Open connection to virtual host."""
        request = pamqp.specification.Connection.Open(
            virtual_host, capabilities, insist
        )
        await self._write_frame(0, request)

    async def open_ok(self, frame):
        self.state = OPEN
        logger.info("Recv open ok")

    #
    ## aioamqp public methods
    #

    async def channel(self, **kwargs):
        """Factory to create a new channel

        """
        await self.ensure_open()
        try:
            channel_id = self.channels_ids_free.pop()
        except KeyError:
            assert self.server_channel_max is not None, 'connection channel-max tuning not performed'
            # channel-max = 0 means no limit
            if self.server_channel_max and self.channels_ids_ceil > self.server_channel_max:
                raise exceptions.NoChannelAvailable()
            self.channels_ids_ceil += 1
            channel_id = self.channels_ids_ceil
        channel = self.CHANNEL_FACTORY(self, channel_id, **kwargs)
        self.channels[channel_id] = channel
        await channel.open()
        return channel



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-four-python.html

"""

import asyncio
import aioamqp

import random
import sys


async def callback(channel, body, envelope, properties):
    print("consumer {} recved {} ({})".format(envelope.consumer_tag, body, envelope.delivery_tag))

async def receive_log(waiter):
    try:
        transport, protocol = await aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    channel = await protocol.channel()
    exchange_name = 'direct_logs'

    await channel.exchange(exchange_name, 'direct')

    result = await channel.queue(queue_name='', durable=False, auto_delete=True)

    queue_name = result['queue']

    severities = sys.argv[1:]
    if not severities:
        print("Usage: %s [info] [warning] [error]" % (sys.argv[0],))
        sys.exit(1)

    for severity in severities:
        await channel.queue_bind(
            exchange_name='direct_logs',
            queue_name=queue_name,
            routing_key=severity,
        )

    print(' [*] Waiting for logs. To exit press CTRL+C')

    await asyncio.wait_for(channel.basic_consume(callback, queue_name=queue_name), timeout=10)
    await waiter.wait()

    await protocol.close()
    transport.close()

loop = asyncio.get_event_loop()

try:
    waiter = asyncio.Event()
    loop.run_until_complete(receive_log(waiter))
except KeyboardInterrupt:
    waiter.set()



import asyncio
import socket
from urllib.parse import urlparse

from .exceptions import *  # pylint: disable=wildcard-import
from .protocol import AmqpProtocol

from .version import __version__
from .version import __packagename__


async def connect(host='localhost', port=None, login='guest', password='guest',
            virtualhost='/', ssl=None, login_method='AMQPLAIN', insist=False,
            protocol_factory=AmqpProtocol, *, loop=None, **kwargs):
    """Convenient method to connect to an AMQP broker

        @host:          the host to connect to
        @port:          broker port
        @login:         login
        @password:      password
        @virtualhost:   AMQP virtualhost to use for this connection
        @ssl:           SSL context used for secure connections, omit for no SSL
                        - see https://docs.python.org/3/library/ssl.html
        @login_method:  AMQP auth method
        @insist:        Insist on connecting to a server
        @protocol_factory:
                        Factory to use, if you need to subclass AmqpProtocol
        @loop:          Set the event loop to use

        @kwargs:        Arguments to be given to the protocol_factory instance

        Returns:        a tuple (transport, protocol) of an AmqpProtocol instance
    """
    if loop is None:
        loop = asyncio.get_event_loop()
    factory = lambda: protocol_factory(loop=loop, **kwargs)

    create_connection_kwargs = {}

    if ssl is not None:
        create_connection_kwargs['ssl'] = ssl

    if port is None:
        if ssl:
            port = 5671
        else:
            port = 5672

    transport, protocol = await loop.create_connection(
        factory, host, port, **create_connection_kwargs
    )

    # these 2 flags *may* show up in sock.type. They are only available on linux
    # see https://bugs.python.org/issue21327
    nonblock = getattr(socket, 'SOCK_NONBLOCK', 0)
    cloexec = getattr(socket, 'SOCK_CLOEXEC', 0)
    sock = transport.get_extra_info('socket')
    if sock is not None and (sock.type & ~nonblock & ~cloexec) == socket.SOCK_STREAM:
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    try:
        await protocol.start_connection(host, port, login, password, virtualhost, ssl=ssl, login_method=login_method,
                                        insist=insist)
    except Exception:
        await protocol.wait_closed()
        raise

    return transport, protocol


async def from_url(
        url, login_method='AMQPLAIN', insist=False, protocol_factory=AmqpProtocol, **kwargs):
    """ Connect to the AMQP using a single url parameter and return the client.

        For instance:

            amqp://user:password@hostname:port/vhost

        @insist:        Insist on connecting to a server
        @protocol_factory:
                        Factory to use, if you need to subclass AmqpProtocol
        @loop:          optionally set the event loop to use.

        @kwargs:        Arguments to be given to the protocol_factory instance

        Returns:        a tuple (transport, protocol) of an AmqpProtocol instance
    """
    url = urlparse(url)

    if url.scheme not in ('amqp', 'amqps'):
        raise ValueError('Invalid protocol %s, valid protocols are amqp or amqps' % url.scheme)

    transport, protocol = await connect(
        host=url.hostname or 'localhost',
        port=url.port,
        login=url.username or 'guest',
        password=url.password or 'guest',
        virtualhost=(url.path[1:] if len(url.path) > 1 else '/'),
        login_method=login_method,
        insist=insist,
        protocol_factory=protocol_factory,
        **kwargs)
    return transport, protocol



import asynctest
import asyncio

from . import testcase


class PublishTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    _multiprocess_can_split_ = True

    async def test_publish(self):
        # declare
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')

        # publish
        await self.channel.publish("coucou", "e", routing_key='')

        queues = self.list_queues()
        self.assertIn("q", queues)
        self.assertEqual(1, queues["q"]['messages'])

    async def test_empty_publish(self):
        # declare
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')

        # publish
        await self.channel.publish("", "e", routing_key='')

        queues = self.list_queues()
        self.assertIn("q", queues)
        self.assertEqual(1, queues["q"]["messages"])
        self.assertEqual(0, queues["q"]["message_bytes"])

    async def test_big_publish(self):
        # declare
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')

        # publish
        await self.channel.publish("a"*1000000, "e", routing_key='')

        queues = self.list_queues()
        self.assertIn("q", queues)
        self.assertEqual(1, queues["q"]['messages'])

    async def test_big_unicode_publish(self):
        # declare
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')

        # publish
        await self.channel.publish("Ы"*1000000, "e", routing_key='')
        await self.channel.publish("Ы"*1000000, "e", routing_key='')

        queues = self.list_queues()
        self.assertIn("q", queues)
        self.assertEqual(2, queues["q"]['messages'])

    async def test_confirmed_publish(self):
        # declare
        await self.channel.confirm_select()
        self.assertTrue(self.channel.publisher_confirms)
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')

        # publish
        await self.channel.publish("coucou", "e", routing_key='')

        queues = self.list_queues()
        self.assertIn("q", queues)
        self.assertEqual(1, queues["q"]['messages'])

    async def test_return_from_publish(self):
        called = False

        async def callback(channel, body, envelope, properties):
            nonlocal called
            called = True
        channel = await self.amqp.channel(return_callback=callback)

        # declare
        await channel.exchange_declare("e", "topic")

        # publish
        await channel.publish("coucou", "e", routing_key="not.found",
                                   mandatory=True)

        for _i in range(10):
            if called:
                break
            await asyncio.sleep(0.1)

        self.assertTrue(called)



#!/usr/bin/env python

"""
    RPC client, aioamqp implementation of RPC examples from RabbitMQ tutorial

"""

import asyncio
import uuid

import aioamqp


class FibonacciRpcClient(object):
    def __init__(self):
        self.transport = None
        self.protocol = None
        self.channel = None
        self.callback_queue = None
        self.waiter = asyncio.Event()

    async def connect(self):
        """ an `__init__` method can't be a coroutine"""
        self.transport, self.protocol = await aioamqp.connect()
        self.channel = await self.protocol.channel()

        result = await self.channel.queue_declare(queue_name='', exclusive=True)
        self.callback_queue = result['queue']

        await self.channel.basic_consume(
            self.on_response,
            no_ack=True,
            queue_name=self.callback_queue,
        )

    async def on_response(self, channel, body, envelope, properties):
        if self.corr_id == properties.correlation_id:
            self.response = body

        self.waiter.set()

    async def call(self, n):
        if not self.protocol:
            await self.connect()
        self.response = None
        self.corr_id = str(uuid.uuid4())
        await self.channel.basic_publish(
            payload=str(n),
            exchange_name='',
            routing_key='rpc_queue',
            properties={
                'reply_to': self.callback_queue,
                'correlation_id': self.corr_id,
            },
        )
        await self.waiter.wait()

        await self.protocol.close()
        return int(self.response)


async def rpc_client():
    fibonacci_rpc = FibonacciRpcClient()
    print(" [x] Requesting fib(30)")
    response = await fibonacci_rpc.call(30)
    print(" [.] Got %r" % response)


asyncio.get_event_loop().run_until_complete(rpc_client())



import asynctest
import asynctest.mock
import asyncio

from aioamqp.protocol import OPEN, CLOSED

from . import testcase


class ConnectionLostTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    _multiprocess_can_split_ = True

    async def test_connection_lost(self):

        self.callback_called = False

        def callback(*args, **kwargs):
            self.callback_called = True

        amqp = self.amqp
        amqp._on_error_callback = callback
        channel = self.channel
        self.assertEqual(amqp.state, OPEN)
        self.assertTrue(channel.is_open)
        amqp._stream_reader._transport.close()  # this should have the same effect as the tcp connection being lost
        await asyncio.wait_for(amqp.worker, 1, loop=self.loop)
        self.assertEqual(amqp.state, CLOSED)
        self.assertFalse(channel.is_open)
        self.assertTrue(self.callback_called)



#!/usr/bin/env python
"""
    Worker example from the 2nd tutorial
"""

import asyncio
import aioamqp

import sys

async def callback(channel, body, envelope, properties):
    print(" [x] Received %r" % body)
    await asyncio.sleep(body.count(b'.'))
    print(" [x] Done")
    await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)


async def worker():
    try:
        transport, protocol = await aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return


    channel = await protocol.channel()

    await channel.queue(queue_name='task_queue', durable=True)
    await channel.basic_qos(prefetch_count=1, prefetch_size=0, connection_global=False)
    await channel.basic_consume(callback, queue_name='task_queue')


event_loop = asyncio.get_event_loop()
event_loop.run_until_complete(worker())
event_loop.run_forever()




import asyncio
import asynctest

from . import testcase
from .. import exceptions
from ..properties import Properties


class ConsumeTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    _multiprocess_can_split_ = True


    def setUp(self):
        super().setUp()
        self.consume_future = asyncio.Future(loop=self.loop)

    async def callback(self, channel, body, envelope, properties):
        self.consume_future.set_result((body, envelope, properties))

    async def get_callback_result(self):
        await self.consume_future
        result = self.consume_future.result()
        self.consume_future = asyncio.Future(loop=self.loop)
        return result

    async def test_consume(self):
        # declare
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')

        # get a different channel
        channel = await self.create_channel()

        # publish
        await channel.publish("coucou", "e", routing_key='',)

        # start consume
        await channel.basic_consume(self.callback, queue_name="q")

        # get one
        body, envelope, properties = await self.get_callback_result()
        self.assertIsNotNone(envelope.consumer_tag)
        self.assertIsNotNone(envelope.delivery_tag)
        self.assertEqual(b"coucou", body)
        self.assertIsInstance(properties, Properties)

    async def test_big_consume(self):
        # declare
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')


        # get a different channel
        channel = await self.create_channel()

        # publish
        await channel.publish("a"*1000000, "e", routing_key='',)

        # start consume
        await channel.basic_consume(self.callback, queue_name="q")

        # get one
        body, envelope, properties = await self.get_callback_result()
        self.assertIsNotNone(envelope.consumer_tag)
        self.assertIsNotNone(envelope.delivery_tag)
        self.assertEqual(b"a"*1000000, body)
        self.assertIsInstance(properties, Properties)

    async def test_consume_multiple_queues(self):
        await self.channel.queue_declare("q1", exclusive=True, no_wait=False)
        await self.channel.queue_declare("q2", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "direct")
        await self.channel.queue_bind("q1", "e", routing_key="q1")
        await self.channel.queue_bind("q2", "e", routing_key="q2")

        # get a different channel
        channel = await self.create_channel()

        q1_future = asyncio.Future(loop=self.loop)

        async def q1_callback(channel, body, envelope, properties):
            q1_future.set_result((body, envelope, properties))

        q2_future = asyncio.Future(loop=self.loop)

        async def q2_callback(channel, body, envelope, properties):
            q2_future.set_result((body, envelope, properties))

        # start consumers
        result = await channel.basic_consume(q1_callback, queue_name="q1")
        ctag_q1 = result['consumer_tag']
        result = await channel.basic_consume(q2_callback, queue_name="q2")
        ctag_q2 = result['consumer_tag']

        # put message in q1
        await channel.publish("coucou1", "e", "q1")

        # get it
        body1, envelope1, properties1 = await q1_future
        self.assertEqual(ctag_q1, envelope1.consumer_tag)
        self.assertIsNotNone(envelope1.delivery_tag)
        self.assertEqual(b"coucou1", body1)
        self.assertIsInstance(properties1, Properties)

        # put message in q2
        await channel.publish("coucou2", "e", "q2")

        # get it
        body2, envelope2, properties2 = await q2_future
        self.assertEqual(ctag_q2, envelope2.consumer_tag)
        self.assertEqual(b"coucou2", body2)
        self.assertIsInstance(properties2, Properties)

    async def test_duplicate_consumer_tag(self):
        await self.channel.queue_declare("q1", exclusive=True, no_wait=False)
        await self.channel.queue_declare("q2", exclusive=True, no_wait=False)
        await self.channel.basic_consume(self.callback, queue_name="q1", consumer_tag='tag')

        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.basic_consume(self.callback, queue_name="q2", consumer_tag='tag')

        self.assertEqual(cm.exception.code, 530)

    async def test_consume_callaback_synced(self):
        # declare
        await self.channel.queue_declare("q", exclusive=True, no_wait=False)
        await self.channel.exchange_declare("e", "fanout")
        await self.channel.queue_bind("q", "e", routing_key='')

        # get a different channel
        channel = await self.create_channel()

        # publish
        await channel.publish("coucou", "e", routing_key='',)

        sync_future = asyncio.Future(loop=self.loop)

        async def callback(channel, body, envelope, properties):
            self.assertTrue(sync_future.done())

        await channel.basic_consume(callback, queue_name="q")
        sync_future.set_result(True)

    async def test_consume_multiple_queues_using_gather(self):
        await asyncio.gather(self.channel.queue_declare("q1", exclusive=True, no_wait=False),
                             self.channel.queue_declare("q2", exclusive=True, no_wait=False))
        await asyncio.gather(self.channel.exchange_declare("e", "direct"),
                             self.channel.exchange_declare("f", "direct"))
        await asyncio.gather(self.channel.queue_bind("q1", "e", routing_key="q1"),
                             self.channel.queue_bind("q2", "e", routing_key="q2"))

        # get a different channel
        channel = await self.create_channel()

        q1_future = asyncio.Future(loop=self.loop)

        async def q1_callback(channel, body, envelope, properties):
            q1_future.set_result((body, envelope, properties))

        q2_future = asyncio.Future(loop=self.loop)

        async def q2_callback(channel, body, envelope, properties):
            q2_future.set_result((body, envelope, properties))

        # start consumers
        results = await asyncio.gather(channel.basic_consume(q1_callback, queue_name="q1"),
                                       channel.basic_consume(q2_callback, queue_name="q2"))
        ctag_q1 = results[0]['consumer_tag']
        ctag_q2 = results[1]['consumer_tag']

        # put message in q1 and q2
        await asyncio.gather(channel.publish("coucou1", "e", "q1"),
                             channel.publish("coucou2", "e", "q2"))

        # get it
        body1, envelope1, properties1 = await q1_future
        self.assertEqual(ctag_q1, envelope1.consumer_tag)
        self.assertIsNotNone(envelope1.delivery_tag)
        self.assertEqual(b"coucou1", body1)
        self.assertIsInstance(properties1, Properties)

        # get it
        body2, envelope2, properties2 = await q2_future
        self.assertEqual(ctag_q2, envelope2.consumer_tag)
        self.assertEqual(b"coucou2", body2)
        self.assertIsInstance(properties2, Properties)



"""
    Hello world `send.py` example implementation using aioamqp.
    See the documentation for more informations.

"""

import asyncio
import aioamqp


async def send():
    transport, protocol = await aioamqp.connect()
    channel = await protocol.channel()

    await channel.queue_declare(queue_name='hello')

    await channel.basic_publish(
        payload='Hello World!',
        exchange_name='',
        routing_key='hello'
    )

    print(" [x] Sent 'Hello World!'")
    await protocol.close()
    transport.close()



asyncio.get_event_loop().run_until_complete(send())



"""
    Server received requests handling tests

"""

import asyncio

import asynctest
import asynctest.mock
import uuid

from . import testcase


async def consumer(channel, body, envelope, properties):
    await channel.basic_client_ack(envelope.delivery_tag)


class ServerBasicCancelTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        super().setUp()
        self.queue_name = str(uuid.uuid4())

    async def test_cancel_whilst_consuming(self):
        await self.channel.queue_declare(self.queue_name)

        # None is non-callable.  We want to make sure the callback is
        # unregistered and never called.
        await self.channel.basic_consume(None)
        await self.channel.queue_delete(self.queue_name)

    async def test_cancel_callbacks(self):
        callback_calls = []
        callback_event = asyncio.Event()

        async def coroutine_callback(*args, **kwargs):
            callback_calls.append((args, kwargs))

        def function_callback(*args, **kwargs):
            callback_calls.append((args, kwargs))
            callback_event.set()

        self.channel.add_cancellation_callback(coroutine_callback)
        self.channel.add_cancellation_callback(function_callback)

        await self.channel.queue_declare(self.queue_name)
        rv = await self.channel.basic_consume(consumer)
        await self.channel.queue_delete(self.queue_name)

        await callback_event.wait()
        self.assertEqual(2, len(callback_calls))
        for args, _kwargs in callback_calls:
            self.assertIs(self.channel, args[0])
            self.assertEqual(rv['consumer_tag'], args[1])

    async def test_cancel_callback_exceptions(self):
        callback_calls = []
        callback_event = asyncio.Event()

        def function_callback(*args, **kwargs):
            callback_calls.append((args, kwargs))
            callback_event.set()
            raise RuntimeError

        self.channel.add_cancellation_callback(function_callback)
        self.channel.add_cancellation_callback(function_callback)

        await self.channel.queue_declare(self.queue_name)
        await self.channel.basic_consume(consumer)
        await self.channel.queue_delete(self.queue_name)

        await callback_event.wait()
        self.assertEqual(2, len(callback_calls))
        self.assertTrue(self.channel.is_open)



"""
    Helper class to decode AMQP responses

AMQP Frame implementations


0      1         3         7                      size+7      size+8
+------+---------+---------+    +-------------+     +-----------+
| type | channel |   size  |    |   payload   |     | frame-end |
+------+---------+---------+    +-------------+     +-----------+
 octets   short     long         'size' octets          octet

The frame-end octet MUST always be the hexadecimal value %xCE

type:

Type = 1, "METHOD": method frame.
Type = 2, "HEADER": content header frame.
Type = 3, "BODY": content body frame.
Type = 4, "HEARTBEAT": heartbeat frame.


Method Payload

0          2           4
+----------+-----------+-------------- - -
| class-id | method-id | arguments...
+----------+-----------+-------------- - -
    short     short       ...

Content Payload

0          2        4           12               14
+----------+--------+-----------+----------------+------------- - -
| class-id | weight | body size | property flags | property list...
+----------+--------+-----------+----------------+------------- - -
   short     short    long long       short        remainder...

"""

import asyncio
import socket

import pamqp.encode
import pamqp.specification
import pamqp.frame

from . import exceptions
from . import constants as amqp_constants


DUMP_FRAMES = False


def write(writer, channel, encoder):
    """Writes the built frame from the encoder

        writer:     asyncio StreamWriter
        channel:    amqp Channel identifier
        encoder:    frame encoder from pamqp which can be marshalled

    Returns int, the number of bytes written.
    """
    return writer.write(pamqp.frame.marshal(encoder, channel))


async def read(reader):
    """Read a new frame from the wire

        reader:     asyncio StreamReader

    Returns (channel, frame) a tuple containing both channel and the pamqp frame,
                             the object describing the frame
    """
    if not reader:
        raise exceptions.AmqpClosedConnection()
    try:
        data = await reader.readexactly(7)
    except (asyncio.IncompleteReadError, socket.error) as ex:
        raise exceptions.AmqpClosedConnection() from ex

    frame_type, channel, frame_length = pamqp.frame.frame_parts(data)

    payload_data = await reader.readexactly(frame_length)
    frame = None

    if frame_type == amqp_constants.TYPE_METHOD:
        frame = pamqp.frame._unmarshal_method_frame(payload_data)

    elif frame_type == amqp_constants.TYPE_HEADER:
        frame = pamqp.frame._unmarshal_header_frame(payload_data)

    elif frame_type == amqp_constants.TYPE_BODY:
        frame = pamqp.frame._unmarshal_body_frame(payload_data)

    elif frame_type == amqp_constants.TYPE_HEARTBEAT:
        frame = pamqp.heartbeat.Heartbeat()

    frame_end = await reader.readexactly(1)
    assert frame_end == amqp_constants.FRAME_END
    return channel, frame



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-five-python.html

"""

import asyncio
import aioamqp

import random
import sys


async def callback(channel, body, envelope, properties):
    print("consumer {} received {} ({})".format(envelope.consumer_tag, body, envelope.delivery_tag))


async def receive_log():
    try:
        transport, protocol = await aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    channel = await protocol.channel()
    exchange_name = 'topic_logs'

    await channel.exchange(exchange_name, 'topic')

    result = await channel.queue(queue_name='', durable=False, auto_delete=True)
    queue_name = result['queue']

    binding_keys = sys.argv[1:]
    if not binding_keys:
        print("Usage: %s [binding_key]..." % (sys.argv[0],))
        sys.exit(1)

    for binding_key in binding_keys:
        await channel.queue_bind(
            exchange_name='topic_logs',
            queue_name=queue_name,
            routing_key=binding_key
        )

    print(' [*] Waiting for logs. To exit press CTRL+C')

    await channel.basic_consume(callback, queue_name=queue_name)

event_loop = asyncio.get_event_loop()
event_loop.run_until_complete(receive_log())
event_loop.run_forever()



"""
    Amqp queue class tests
"""

import asyncio
import asynctest

from . import testcase
from .. import exceptions


class QueueDeclareTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    def setUp(self):
        super().setUp()
        self.consume_future = asyncio.Future(loop=self.loop)

    async def callback(self, body, envelope, properties):
        self.consume_future.set_result((body, envelope, properties))

    async def get_callback_result(self):
        await self.consume_future
        result = self.consume_future.result()
        self.consume_future = asyncio.Future(loop=self.loop)
        return result

    async def test_queue_declare_no_name(self):
        result = await self.channel.queue_declare()
        self.assertIsNotNone(result['queue'])

    async def test_queue_declare(self):
        queue_name = 'queue_name'
        result = await self.channel.queue_declare('queue_name')
        self.assertEqual(result['message_count'], 0)
        self.assertEqual(result['consumer_count'], 0)
        self.assertEqual(result['queue'].split('.')[-1], queue_name)
        self.assertTrue(result)

    async def test_queue_declare_passive(self):
        queue_name = 'queue_name'
        await self.channel.queue_declare('queue_name')
        result = await self.channel.queue_declare(queue_name, passive=True)
        self.assertEqual(result['message_count'], 0)
        self.assertEqual(result['consumer_count'], 0)
        self.assertEqual(result['queue'].split('.')[-1], queue_name)

    async def test_queue_declare_custom_x_message_ttl_32_bits(self):
        queue_name = 'queue_name'
        # 2147483648 == 10000000000000000000000000000000
        # in binary, meaning it is 32 bit long
        x_message_ttl = 2147483648
        result = await self.channel.queue_declare('queue_name', arguments={
            'x-message-ttl': x_message_ttl
        })
        self.assertEqual(result['message_count'], 0)
        self.assertEqual(result['consumer_count'], 0)
        self.assertEqual(result['queue'].split('.')[-1], queue_name)
        self.assertTrue(result)

    async def test_queue_declare_passive_nonexistant_queue(self):
        queue_name = 'queue_name'
        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.queue_declare(queue_name, passive=True)

        self.assertEqual(cm.exception.code, 404)

    async def test_wrong_parameter_queue(self):
        queue_name = 'queue_name'
        await self.channel.queue_declare(queue_name, exclusive=False, auto_delete=False)

        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.queue_declare(queue_name,
                passive=False, exclusive=True, auto_delete=True)

        self.assertIn(cm.exception.code, [405, 406])

    async def test_multiple_channel_same_queue(self):
        queue_name = 'queue_name'

        channel1 = await self.amqp.channel()
        channel2 = await self.amqp.channel()

        result = await channel1.queue_declare(queue_name, passive=False)
        self.assertEqual(result['message_count'], 0)
        self.assertEqual(result['consumer_count'], 0)
        self.assertEqual(result['queue'].split('.')[-1], queue_name)

        result = await channel2.queue_declare(queue_name, passive=False)
        self.assertEqual(result['message_count'], 0)
        self.assertEqual(result['consumer_count'], 0)
        self.assertEqual(result['queue'].split('.')[-1], queue_name)

    async def _test_queue_declare(self, queue_name, exclusive=False, durable=False, auto_delete=False):
        # declare queue
        result = await self.channel.queue_declare(
            queue_name, no_wait=False, exclusive=exclusive, durable=durable,
            auto_delete=auto_delete)

        # assert returned results has the good arguments
        # in test the channel declared queues with prefixed names, to get the full name of the
        # declared queue we have to use self.full_name function
        self.assertEqual(self.full_name(queue_name), result['queue'])

        queues = self.list_queues()
        queue = queues[queue_name]

        # assert queue has been declared witht the good arguments
        self.assertEqual(queue_name, queue['name'])
        self.assertEqual(0, queue['consumers'])
        self.assertEqual(0, queue['messages_ready'])
        self.assertEqual(auto_delete, queue['auto_delete'])
        self.assertEqual(durable, queue['durable'])

        # delete queue
        await self.safe_queue_delete(queue_name)

    def test_durable_and_auto_deleted(self):
        self.loop.run_until_complete(
            self._test_queue_declare('q', exclusive=False, durable=True, auto_delete=True))

    def test_durable_and_not_auto_deleted(self):
        self.loop.run_until_complete(
            self._test_queue_declare('q', exclusive=False, durable=True, auto_delete=False))

    def test_not_durable_and_auto_deleted(self):
        self.loop.run_until_complete(
            self._test_queue_declare('q', exclusive=False, durable=False, auto_delete=True))

    def test_not_durable_and_not_auto_deleted(self):
        self.loop.run_until_complete(
            self._test_queue_declare('q', exclusive=False, durable=False, auto_delete=False))

    async def test_exclusive(self):
        # create an exclusive queue
        await self.channel.queue_declare("q", exclusive=True)
        # consume it
        await self.channel.basic_consume(self.callback, queue_name="q", no_wait=False)
        # create an other amqp connection

        _transport, protocol = await self.create_amqp()
        channel = await self.create_channel(amqp=protocol)
        # assert that this connection cannot connect to the queue
        with self.assertRaises(exceptions.ChannelClosed):
            await channel.basic_consume(self.callback, queue_name="q", no_wait=False)
        # amqp and channels are auto deleted by test case

    async def test_not_exclusive(self):
        # create a non-exclusive queue
        await self.channel.queue_declare('q', exclusive=False)
        # consume it
        await self.channel.basic_consume(self.callback, queue_name='q', no_wait=False)
        # create an other amqp connection
        _transport, protocol = await self.create_amqp()
        channel = await self.create_channel(amqp=protocol)
        # assert that this connection can connect to the queue
        await channel.basic_consume(self.callback, queue_name='q', no_wait=False)


class QueueDeleteTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):


    async def test_delete_queue(self):
        queue_name = 'queue_name'
        await self.channel.queue_declare(queue_name)
        result = await self.channel.queue_delete(queue_name)
        self.assertTrue(result)

    async def test_delete_inexistant_queue(self):
        queue_name = 'queue_name'
        if self.server_version() < (3, 3, 5):
            with self.assertRaises(exceptions.ChannelClosed) as cm:
                result = await self.channel.queue_delete(queue_name)

            self.assertEqual(cm.exception.code, 404)

        else:
            result = await self.channel.queue_delete(queue_name)
            self.assertTrue(result)

class QueueBindTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):


    async def test_bind_queue(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'

        await self.channel.queue_declare(queue_name)
        await self.channel.exchange_declare(exchange_name, type_name='direct')

        result = await self.channel.queue_bind(queue_name, exchange_name, routing_key='')
        self.assertTrue(result)

    async def test_bind_unexistant_exchange(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'


        await self.channel.queue_declare(queue_name)
        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.queue_bind(queue_name, exchange_name, routing_key='')
        self.assertEqual(cm.exception.code, 404)

    async def test_bind_unexistant_queue(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'


        await self.channel.exchange_declare(exchange_name, type_name='direct')

        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.queue_bind(queue_name, exchange_name, routing_key='')
        self.assertEqual(cm.exception.code, 404)

    async def test_unbind_queue(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'

        await self.channel.queue_declare(queue_name)
        await self.channel.exchange_declare(exchange_name, type_name='direct')

        await self.channel.queue_bind(queue_name, exchange_name, routing_key='')

        result = await self.channel.queue_unbind(queue_name, exchange_name, routing_key='')
        self.assertTrue(result)


class QueuePurgeTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):


    async def test_purge_queue(self):
        queue_name = 'queue_name'

        await self.channel.queue_declare(queue_name)
        result = await self.channel.queue_purge(queue_name)
        self.assertEqual(result['message_count'], 0)

    async def test_purge_queue_inexistant_queue(self):
        queue_name = 'queue_name'

        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.queue_purge(queue_name)
        self.assertEqual(cm.exception.code, 404)



"""
    RPC server, aioamqp implementation of RPC examples from RabbitMQ tutorial
"""

import asyncio
import aioamqp


def fib(n):
    if n == 0:
        return 0
    elif n == 1:
        return 1
    else:
        return fib(n-1) + fib(n-2)


async def on_request(channel, body, envelope, properties):
    n = int(body)

    print(" [.] fib(%s)" % n)
    response = fib(n)

    await channel.basic_publish(
        payload=str(response),
        exchange_name='',
        routing_key=properties.reply_to,
        properties={
            'correlation_id': properties.correlation_id,
        },
    )

    await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)


async def rpc_server():

    transport, protocol = await aioamqp.connect()

    channel = await protocol.channel()

    await channel.queue_declare(queue_name='rpc_queue')
    await channel.basic_qos(prefetch_count=1, prefetch_size=0, connection_global=False)
    await channel.basic_consume(on_request, queue_name='rpc_queue')
    print(" [x] Awaiting RPC requests")


event_loop = asyncio.get_event_loop()
event_loop.run_until_complete(rpc_server())
event_loop.run_forever()



"""Aioamqp tests"""

import asynctest
import socket

from aioamqp import connect
from aioamqp.protocol import OPEN

from . import testcase


class AmqpConnectionTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_connect(self):
        _transport, proto = await connect(host=self.host, port=self.port, virtualhost=self.vhost, loop=self.loop)
        self.assertEqual(proto.state, OPEN)
        self.assertIsNotNone(proto.server_properties)
        await proto.close()

    async def test_connect_tuning(self):
        # frame_max should be higher than 131072
        frame_max = 131072
        channel_max = 10
        heartbeat = 100
        _transport, proto = await connect(
            host=self.host,
            port=self.port,
            virtualhost=self.vhost,
            loop=self.loop,
            channel_max=channel_max,
            frame_max=frame_max,
            heartbeat=heartbeat,
        )
        self.assertEqual(proto.state, OPEN)
        self.assertIsNotNone(proto.server_properties)

        self.assertDictEqual(proto.connection_tunning, {
            'frame_max': frame_max,
            'channel_max': channel_max,
            'heartbeat': heartbeat
        })

        self.assertEqual(proto.server_channel_max, channel_max)
        self.assertEqual(proto.server_frame_max, frame_max)
        self.assertEqual(proto.server_heartbeat, heartbeat)

        await proto.close()

    async def test_socket_nodelay(self):
        transport, proto = await connect(host=self.host, port=self.port, virtualhost=self.vhost, loop=self.loop)
        sock = transport.get_extra_info('socket')
        opt_val = sock.getsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY)
        self.assertNotEqual(opt_val, 0)
        await proto.close()



"""
    Amqp basic tests for recover methods
"""

import asynctest

from . import testcase


class RecoverTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_basic_recover_async(self):
        await self.channel.basic_recover_async(requeue=True)

    async def test_basic_recover_async_no_requeue(self):
        await self.channel.basic_recover_async(requeue=False)

    async def test_basic_recover(self):
        result = await self.channel.basic_recover(requeue=True)
        self.assertTrue(result)



"""
    Test our Protocol class
"""
import asynctest
from unittest import mock

from . import testcase
from .. import exceptions
from .. import connect as amqp_connect
from .. import from_url as amqp_from_url
from ..protocol import AmqpProtocol, OPEN


class ProtocolTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_connect(self):
        _transport, protocol = await amqp_connect(
            host=self.host, port=self.port, virtualhost=self.vhost, loop=self.loop
        )
        self.assertEqual(protocol.state, OPEN)
        await protocol.close()

    async def test_connect_products_info(self):
        client_properties = {
            'program': 'aioamqp-tests',
            'program_version': '0.1.1',
        }
        _transport, protocol = await amqp_connect(
            host=self.host,
            port=self.port,
            virtualhost=self.vhost,
            client_properties=client_properties,
            loop=self.loop,
        )

        self.assertEqual(protocol.client_properties, client_properties)
        await protocol.close()

    async def test_connection_unexistant_vhost(self):
        with self.assertRaises(exceptions.AmqpClosedConnection):
            await amqp_connect(host=self.host, port=self.port, virtualhost='/unexistant', loop=self.loop)

    def test_connection_wrong_login_password(self):
        with self.assertRaises(exceptions.AmqpClosedConnection):
            self.loop.run_until_complete(
                amqp_connect(host=self.host, port=self.port, login='wrong', password='wrong', loop=self.loop)
            )

    async def test_connection_from_url(self):
        with mock.patch('aioamqp.connect') as connect:
            async def func(*x, **y):
                return 1, 2
            connect.side_effect = func
            await amqp_from_url('amqp://tom:pass@example.com:7777/myvhost', loop=self.loop)
            connect.assert_called_once_with(
                insist=False,
                password='pass',
                login_method='AMQPLAIN',
                login='tom',
                host='example.com',
                protocol_factory=AmqpProtocol,
                virtualhost='myvhost',
                port=7777,
                loop=self.loop,
            )

    async def test_ssl_context_connection_from_url(self):
        ssl_context = mock.Mock()
        with mock.patch('aioamqp.connect') as connect:
            async def func(*x, **y):
                return 1, 2
            connect.side_effect = func
            await amqp_from_url('amqps://tom:pass@example.com:7777/myvhost', loop=self.loop, ssl=ssl_context)
            connect.assert_called_once_with(
                insist=False,
                password='pass',
                login_method='AMQPLAIN',
                ssl=ssl_context,
                login='tom',
                host='example.com',
                protocol_factory=AmqpProtocol,
                virtualhost='myvhost',
                port=7777,
                loop=self.loop,
            )

    async def test_from_url_raises_on_wrong_scheme(self):
        with self.assertRaises(ValueError):
            await amqp_from_url('invalid://')



"""
    Amqp channel specification
"""

import asyncio
import logging
import uuid
import io
from itertools import count
import warnings

import pamqp.specification

from . import frame as amqp_frame
from . import exceptions
from . import properties as amqp_properties
from .envelope import Envelope, ReturnEnvelope


logger = logging.getLogger(__name__)


class Channel:

    def __init__(self, protocol, channel_id, return_callback=None):
        self._loop = protocol._loop
        self.protocol = protocol
        self.channel_id = channel_id
        self.consumer_queues = {}
        self.consumer_callbacks = {}
        self.cancellation_callbacks = []
        self.return_callback = return_callback
        self.response_future = None
        self.close_event = asyncio.Event(loop=self._loop)
        self.cancelled_consumers = set()
        self.last_consumer_tag = None
        self.publisher_confirms = False
        self.delivery_tag_iter = None  # used for mapping delivered messages to publisher confirms

        self._exchange_declare_lock = asyncio.Lock()
        self._queue_bind_lock = asyncio.Lock()
        self._futures = {}
        self._ctag_events = {}

    def _set_waiter(self, rpc_name):
        if rpc_name in self._futures:
            raise exceptions.SynchronizationError("Waiter already exists")

        fut = asyncio.Future(loop=self._loop)
        self._futures[rpc_name] = fut
        return fut

    def _get_waiter(self, rpc_name):
        fut = self._futures.pop(rpc_name, None)
        if not fut:
            raise exceptions.SynchronizationError("Call %s didn't set a waiter" % rpc_name)
        return fut

    @property
    def is_open(self):
        return not self.close_event.is_set()

    def connection_closed(self, server_code=None, server_reason=None, exception=None):
        for future in self._futures.values():
            if future.done():
                continue
            if exception is None:
                kwargs = {}
                if server_code is not None:
                    kwargs['code'] = server_code
                if server_reason is not None:
                    kwargs['message'] = server_reason
                exception = exceptions.ChannelClosed(**kwargs)
            future.set_exception(exception)

        self.protocol.release_channel_id(self.channel_id)
        self.close_event.set()

    async def dispatch_frame(self, frame):
        methods = {
            pamqp.specification.Channel.OpenOk.name: self.open_ok,
            pamqp.specification.Channel.FlowOk.name: self.flow_ok,
            pamqp.specification.Channel.CloseOk.name: self.close_ok,
            pamqp.specification.Channel.Close.name: self.server_channel_close,

            pamqp.specification.Exchange.DeclareOk.name: self.exchange_declare_ok,
            pamqp.specification.Exchange.BindOk.name: self.exchange_bind_ok,
            pamqp.specification.Exchange.UnbindOk.name: self.exchange_unbind_ok,
            pamqp.specification.Exchange.DeleteOk.name: self.exchange_delete_ok,

            pamqp.specification.Queue.DeclareOk.name: self.queue_declare_ok,
            pamqp.specification.Queue.DeleteOk.name: self.queue_delete_ok,
            pamqp.specification.Queue.BindOk.name: self.queue_bind_ok,
            pamqp.specification.Queue.UnbindOk.name: self.queue_unbind_ok,
            pamqp.specification.Queue.PurgeOk.name: self.queue_purge_ok,

            pamqp.specification.Basic.QosOk.name: self.basic_qos_ok,
            pamqp.specification.Basic.ConsumeOk.name: self.basic_consume_ok,
            pamqp.specification.Basic.CancelOk.name: self.basic_cancel_ok,
            pamqp.specification.Basic.GetOk.name: self.basic_get_ok,
            pamqp.specification.Basic.GetEmpty.name: self.basic_get_empty,
            pamqp.specification.Basic.Deliver.name: self.basic_deliver,
            pamqp.specification.Basic.Cancel.name: self.server_basic_cancel,
            pamqp.specification.Basic.Ack.name: self.basic_server_ack,
            pamqp.specification.Basic.Nack.name: self.basic_server_nack,
            pamqp.specification.Basic.RecoverOk.name: self.basic_recover_ok,
            pamqp.specification.Basic.Return.name: self.basic_return,

            pamqp.specification.Confirm.SelectOk.name: self.confirm_select_ok,
        }

        if frame.name not in methods:
            raise NotImplementedError("Frame %s is not implemented" % frame.name)

        await methods[frame.name](frame)

    async def _write_frame(self, channel_id, request, check_open=True, drain=True):
        await self.protocol.ensure_open()
        if not self.is_open and check_open:
            raise exceptions.ChannelClosed()
        amqp_frame.write(self.protocol._stream_writer, channel_id, request)
        if drain:
            await self.protocol._drain()

    async def _write_frame_awaiting_response(self, waiter_id, channel_id, request,
                                             no_wait, check_open=True, drain=True):
        '''Write a frame and set a waiter for the response (unless no_wait is set)'''
        if no_wait:
            await self._write_frame(channel_id, request, check_open=check_open, drain=drain)
            return None

        f = self._set_waiter(waiter_id)
        try:
            await self._write_frame(channel_id, request, check_open=check_open, drain=drain)
        except Exception:
            self._get_waiter(waiter_id)
            f.cancel()
            raise
        return (await f)

#
## Channel class implementation
#

    async def open(self):
        """Open the channel on the server."""
        request = pamqp.specification.Channel.Open()
        return (await self._write_frame_awaiting_response(
            'open', self.channel_id, request, no_wait=False, check_open=False))

    async def open_ok(self, frame):
        self.close_event.clear()
        fut = self._get_waiter('open')
        fut.set_result(True)
        logger.debug("Channel is open")

    async def close(self, reply_code=0, reply_text="Normal Shutdown"):
        """Close the channel."""
        if not self.is_open:
            raise exceptions.ChannelClosed("channel already closed or closing")
        self.close_event.set()
        request = pamqp.specification.Channel.Close(reply_code, reply_text, class_id=0, method_id=0)
        return (await self._write_frame_awaiting_response(
            'close', self.channel_id, request, no_wait=False, check_open=False))

    async def close_ok(self, frame):
        self._get_waiter('close').set_result(True)
        logger.info("Channel closed")
        self.protocol.release_channel_id(self.channel_id)

    async def _send_channel_close_ok(self):
        request = pamqp.specification.Channel.CloseOk()
        await self._write_frame(self.channel_id, request)

    async def server_channel_close(self, frame):
        await self._send_channel_close_ok()
        results = {
            'reply_code': frame.reply_code,
            'reply_text': frame.reply_text,
            'class_id': frame.class_id,
            'method_id': frame.method_id,
        }
        self.connection_closed(results['reply_code'], results['reply_text'])

    async def flow(self, active):
        request = pamqp.specification.Channel.Flow(active)
        return (await self._write_frame_awaiting_response(
            'flow', self.channel_id, request, no_wait=False,
            check_open=False))

    async def flow_ok(self, frame):
        self.close_event.clear()
        fut = self._get_waiter('flow')
        fut.set_result({'active': frame.active})

        logger.debug("Flow ok")

#
## Exchange class implementation
#

    async def exchange_declare(self, exchange_name, type_name, passive=False, durable=False,
                         auto_delete=False, no_wait=False, arguments=None):
        request = pamqp.specification.Exchange.Declare(
            exchange=exchange_name,
            exchange_type=type_name,
            passive=passive,
            durable=durable,
            auto_delete=auto_delete,
            nowait=no_wait,
            arguments=arguments
        )

        async with self._exchange_declare_lock:
            return (await self._write_frame_awaiting_response(
                'exchange_declare', self.channel_id, request, no_wait))

    async def exchange_declare_ok(self, frame):
        future = self._get_waiter('exchange_declare')
        future.set_result(True)
        logger.debug("Exchange declared")
        return future

    async def exchange_delete(self, exchange_name, if_unused=False, no_wait=False):
        request = pamqp.specification.Exchange.Delete(exchange=exchange_name, if_unused=if_unused, nowait=no_wait)
        return await self._write_frame_awaiting_response(
            'exchange_delete', self.channel_id, request, no_wait)

    async def exchange_delete_ok(self, frame):
        future = self._get_waiter('exchange_delete')
        future.set_result(True)
        logger.debug("Exchange deleted")

    async def exchange_bind(self, exchange_destination, exchange_source, routing_key,
                      no_wait=False, arguments=None):
        if arguments is None:
            arguments = {}
        request = pamqp.specification.Exchange.Bind(
            destination=exchange_destination,
            source=exchange_source,
            routing_key=routing_key,
            nowait=no_wait,
            arguments=arguments
        )
        return (await self._write_frame_awaiting_response(
            'exchange_bind', self.channel_id, request, no_wait))

    async def exchange_bind_ok(self, frame):
        future = self._get_waiter('exchange_bind')
        future.set_result(True)
        logger.debug("Exchange bound")

    async def exchange_unbind(self, exchange_destination, exchange_source, routing_key,
                        no_wait=False, arguments=None):
        if arguments is None:
            arguments = {}

        request = pamqp.specification.Exchange.Unbind(
            destination=exchange_destination,
            source=exchange_source,
            routing_key=routing_key,
            nowait=no_wait,
            arguments=arguments,
        )
        return (await self._write_frame_awaiting_response(
            'exchange_unbind', self.channel_id, request, no_wait))

    async def exchange_unbind_ok(self, frame):
        future = self._get_waiter('exchange_unbind')
        future.set_result(True)
        logger.debug("Exchange bound")

#
## Queue class implementation
#

    async def queue_declare(self, queue_name=None, passive=False, durable=False,
                      exclusive=False, auto_delete=False, no_wait=False, arguments=None):
        """Create or check a queue on the broker
           Args:
               queue_name:     str, the queue to receive message from.
                               The server generate a queue_name if not specified.
               passive:        bool, if set, the server will reply with
                               Declare-Ok if the queue already exists with the same name, and
                               raise an error if not. Checks for the same parameter as well.
               durable:        bool: If set when creating a new queue, the queue
                               will be marked as durable. Durable queues remain active when a
               server restarts.
               exclusive:      bool, request exclusive consumer access,
                               meaning only this consumer can access the queue
               no_wait:        bool, if set, the server will not respond to the method
               arguments:      dict, AMQP arguments to be passed when creating
               the queue.
        """
        if arguments is None:
            arguments = {}

        if not queue_name:
            queue_name = 'aioamqp.gen-' + str(uuid.uuid4())
        request = pamqp.specification.Queue.Declare(
            queue=queue_name,
            passive=passive,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            nowait=no_wait,
            arguments=arguments
        )
        return (await self._write_frame_awaiting_response(
            'queue_declare' + queue_name, self.channel_id, request, no_wait))

    async def queue_declare_ok(self, frame):
        results = {
            'queue': frame.queue,
            'message_count': frame.message_count,
            'consumer_count': frame.consumer_count,
        }
        future = self._get_waiter('queue_declare' + results['queue'])
        future.set_result(results)
        logger.debug("Queue declared")

    async def queue_delete(self, queue_name, if_unused=False, if_empty=False, no_wait=False):
        """Delete a queue in RabbitMQ
            Args:
               queue_name:     str, the queue to receive message from
               if_unused:      bool, the queue is deleted if it has no consumers. Raise if not.
               if_empty:       bool, the queue is deleted if it has no messages. Raise if not.
               no_wait:        bool, if set, the server will not respond to the method
        """
        request = pamqp.specification.Queue.Delete(
            queue=queue_name,
            if_unused=if_unused,
            if_empty=if_empty,
            nowait=no_wait
        )
        return (await self._write_frame_awaiting_response(
            'queue_delete', self.channel_id, request, no_wait))

    async def queue_delete_ok(self, frame):
        future = self._get_waiter('queue_delete')
        future.set_result(True)
        logger.debug("Queue deleted")

    async def queue_bind(self, queue_name, exchange_name, routing_key, no_wait=False, arguments=None):
        """Bind a queue and a channel."""
        if arguments is None:
            arguments = {}

        request = pamqp.specification.Queue.Bind(
            queue=queue_name,
            exchange=exchange_name,
            routing_key=routing_key,
            nowait=no_wait,
            arguments=arguments
        )
        # short reserved-1
        async with self._queue_bind_lock:
            return (await self._write_frame_awaiting_response(
                'queue_bind', self.channel_id, request, no_wait))

    async def queue_bind_ok(self, frame):
        future = self._get_waiter('queue_bind')
        future.set_result(True)
        logger.debug("Queue bound")

    async def queue_unbind(self, queue_name, exchange_name, routing_key, arguments=None):
        if arguments is None:
            arguments = {}

        request = pamqp.specification.Queue.Unbind(
            queue=queue_name,
            exchange=exchange_name,
            routing_key=routing_key,
            arguments=arguments
        )

        return (await self._write_frame_awaiting_response(
            'queue_unbind', self.channel_id, request, no_wait=False))

    async def queue_unbind_ok(self, frame):
        future = self._get_waiter('queue_unbind')
        future.set_result(True)
        logger.debug("Queue unbound")

    async def queue_purge(self, queue_name, no_wait=False):
        request = pamqp.specification.Queue.Purge(
            queue=queue_name, nowait=no_wait
        )
        return (await self._write_frame_awaiting_response(
            'queue_purge', self.channel_id, request, no_wait=no_wait))

    async def queue_purge_ok(self, frame):
        future = self._get_waiter('queue_purge')
        future.set_result({'message_count': frame.message_count})

#
## Basic class implementation
#

    async def basic_publish(self, payload, exchange_name, routing_key,
                            properties=None, mandatory=False, immediate=False):
        if isinstance(payload, str):
            warnings.warn("Str payload support will be removed in next release", DeprecationWarning)
            payload = payload.encode()

        if properties is None:
            properties = {}

        method_request = pamqp.specification.Basic.Publish(
            exchange=exchange_name,
            routing_key=routing_key,
            mandatory=mandatory,
            immediate=immediate
        )

        await self._write_frame(self.channel_id, method_request, drain=False)

        header_request = pamqp.header.ContentHeader(
            body_size=len(payload),
            properties=pamqp.specification.Basic.Properties(**properties)
        )
        await self._write_frame(self.channel_id, header_request, drain=False)

        # split the payload

        frame_max = self.protocol.server_frame_max or len(payload)
        for chunk in (payload[0+i:frame_max+i] for i in range(0, len(payload), frame_max)):
            content_request = pamqp.body.ContentBody(chunk)
            await self._write_frame(self.channel_id, content_request, drain=False)

        await self.protocol._drain()

    async def basic_qos(self, prefetch_size=0, prefetch_count=0, connection_global=False):
        """Specifies quality of service.

        Args:
            prefetch_size:      int, request that messages be sent in advance so
                                that when the client finishes processing a message, the
                                following message is already held locally
            prefetch_count:     int: Specifies a prefetch window in terms of
                                whole messages. This field may be used in combination with the
                                prefetch-size field; a message will only be sent in advance if
                                both prefetch windows (and those at the channel and connection
                                level) allow it
            connection_global:  bool: global=false means that the QoS
                                settings should apply per-consumer channel; and global=true to mean
                                that the QoS settings should apply per-channel.
        """
        request = pamqp.specification.Basic.Qos(
            prefetch_size, prefetch_count, connection_global
        )
        return (await self._write_frame_awaiting_response(
            'basic_qos', self.channel_id, request, no_wait=False)
        )

    async def basic_qos_ok(self, frame):
        future = self._get_waiter('basic_qos')
        future.set_result(True)
        logger.debug("Qos ok")


    async def basic_server_nack(self, frame, delivery_tag=None):
        if delivery_tag is None:
            delivery_tag = frame.delivery_tag
        fut = self._get_waiter('basic_server_ack_{}'.format(delivery_tag))
        logger.debug('Received nack for delivery tag %r', delivery_tag)
        fut.set_exception(exceptions.PublishFailed(delivery_tag))

    async def basic_consume(self, callback, queue_name='', consumer_tag='', no_local=False, no_ack=False,
                      exclusive=False, no_wait=False, arguments=None):
        """Starts the consumption of message into a queue.
        the callback will be called each time we're receiving a message.

            Args:
                callback:       coroutine, the called callback
                queue_name:     str, the queue to receive message from
                consumer_tag:   str, optional consumer tag
                no_local:       bool, if set the server will not send messages
                                to the connection that published them.
                no_ack:         bool, if set the server does not expect
                                acknowledgements for messages
                exclusive:      bool, request exclusive consumer access,
                                meaning only this consumer can access the queue
                no_wait:        bool, if set, the server will not respond to the method
                arguments:      dict, AMQP arguments to be passed to the server
        """
        # If a consumer tag was not passed, create one
        consumer_tag = consumer_tag or 'ctag%i.%s' % (self.channel_id, uuid.uuid4().hex)

        if arguments is None:
            arguments = {}

        request = pamqp.specification.Basic.Consume(
            queue=queue_name,
            consumer_tag=consumer_tag,
            no_local=no_local,
            no_ack=no_ack,
            exclusive=exclusive,
            nowait=no_wait,
            arguments=arguments
        )

        self.consumer_callbacks[consumer_tag] = callback
        self.last_consumer_tag = consumer_tag

        return_value = await self._write_frame_awaiting_response(
            'basic_consume' + consumer_tag, self.channel_id, request, no_wait)
        if no_wait:
            return_value = {'consumer_tag': consumer_tag}
        else:
            self._ctag_events[consumer_tag].set()
        return return_value

    async def basic_consume_ok(self, frame):
        ctag = frame.consumer_tag
        results = {
            'consumer_tag': ctag,
        }
        future = self._get_waiter('basic_consume' + ctag)
        future.set_result(results)
        self._ctag_events[ctag] = asyncio.Event(loop=self._loop)

    async def basic_deliver(self, frame):
        consumer_tag = frame.consumer_tag
        delivery_tag = frame.delivery_tag
        is_redeliver = frame.redelivered
        exchange_name = frame.exchange
        routing_key = frame.routing_key
        _channel, content_header_frame = await self.protocol.get_frame()

        buffer = io.BytesIO()

        while(buffer.tell() < content_header_frame.body_size):
            _channel, content_body_frame = await self.protocol.get_frame()
            buffer.write(content_body_frame.value)

        body = buffer.getvalue()
        envelope = Envelope(consumer_tag, delivery_tag, exchange_name, routing_key, is_redeliver)
        properties = amqp_properties.from_pamqp(content_header_frame.properties)

        callback = self.consumer_callbacks[consumer_tag]

        event = self._ctag_events.get(consumer_tag)
        if event:
            await event.wait()
            del self._ctag_events[consumer_tag]

        await callback(self, body, envelope, properties)

    async def server_basic_cancel(self, frame):
        # https://www.rabbitmq.com/consumer-cancel.html
        consumer_tag = frame.consumer_tag
        _no_wait = frame.nowait
        self.cancelled_consumers.add(consumer_tag)
        logger.info("consume cancelled received")
        for callback in self.cancellation_callbacks:
            try:
                await callback(self, consumer_tag)
            except Exception as error:  # pylint: disable=broad-except
                logger.error("cancellation callback %r raised exception %r",
                             callback, error)

    async def basic_cancel(self, consumer_tag, no_wait=False):
        request = pamqp.specification.Basic.Cancel(consumer_tag, no_wait)
        return (await self._write_frame_awaiting_response(
            'basic_cancel', self.channel_id, request, no_wait=no_wait)
        )

    async def basic_cancel_ok(self, frame):
        results = {
            'consumer_tag': frame.consumer_tag,
        }
        future = self._get_waiter('basic_cancel')
        future.set_result(results)
        logger.debug("Cancel ok")

    async def basic_get(self, queue_name='', no_ack=False):
        request = pamqp.specification.Basic.Get(queue=queue_name, no_ack=no_ack)
        return (await self._write_frame_awaiting_response(
            'basic_get', self.channel_id, request, no_wait=False)
        )

    async def basic_get_ok(self, frame):
        data = {
            'delivery_tag': frame.delivery_tag,
            'redelivered': frame.redelivered,
            'exchange_name': frame.exchange,
            'routing_key': frame.routing_key,
            'message_count': frame.message_count,
        }

        _channel, content_header_frame = await self.protocol.get_frame()

        buffer = io.BytesIO()
        while(buffer.tell() < content_header_frame.body_size):
            _channel, content_body_frame = await self.protocol.get_frame()
            buffer.write(content_body_frame.value)

        data['message'] = buffer.getvalue()
        data['properties'] = amqp_properties.from_pamqp(content_header_frame.properties)
        future = self._get_waiter('basic_get')
        future.set_result(data)

    async def basic_get_empty(self, frame):
        future = self._get_waiter('basic_get')
        future.set_exception(exceptions.EmptyQueue)

    async def basic_client_ack(self, delivery_tag, multiple=False):
        request = pamqp.specification.Basic.Ack(delivery_tag, multiple)
        await self._write_frame(self.channel_id, request)

    async def basic_client_nack(self, delivery_tag, multiple=False, requeue=True):
        request = pamqp.specification.Basic.Nack(delivery_tag, multiple, requeue)
        await self._write_frame(self.channel_id, request)

    async def basic_server_ack(self, frame):
        delivery_tag = frame.delivery_tag
        fut = self._get_waiter('basic_server_ack_{}'.format(delivery_tag))
        logger.debug('Received ack for delivery tag %s', delivery_tag)
        fut.set_result(True)

    async def basic_reject(self, delivery_tag, requeue=False):
        request = pamqp.specification.Basic.Reject(delivery_tag, requeue)
        await self._write_frame(self.channel_id, request)

    async def basic_recover_async(self, requeue=True):
        request = pamqp.specification.Basic.RecoverAsync(requeue)
        await self._write_frame(self.channel_id, request)

    async def basic_recover(self, requeue=True):
        request = pamqp.specification.Basic.Recover(requeue)
        return (await self._write_frame_awaiting_response(
            'basic_recover', self.channel_id, request, no_wait=False)
        )

    async def basic_recover_ok(self, frame):
        future = self._get_waiter('basic_recover')
        future.set_result(True)
        logger.debug("Cancel ok")

    async def basic_return(self, frame):
        reply_code = frame.reply_code
        reply_text = frame.reply_text
        exchange_name = frame.exchange
        routing_key = frame.routing_key
        _channel, content_header_frame = await self.protocol.get_frame()

        buffer = io.BytesIO()
        while(buffer.tell() < content_header_frame.body_size):
            _channel, content_body_frame = await self.protocol.get_frame()
            buffer.write(content_body_frame.value)

        body = buffer.getvalue()
        envelope = ReturnEnvelope(reply_code, reply_text,
                                  exchange_name, routing_key)
        properties = amqp_properties.from_pamqp(content_header_frame.properties)
        callback = self.return_callback
        if callback is None:
            # they have set mandatory bit, but havent added a callback
            logger.warning('You have received a returned message, but dont have a callback registered for returns.'
                           ' Please set channel.return_callback')
        else:
            await callback(self, body, envelope, properties)


#
## convenient aliases
#
    queue = queue_declare
    exchange = exchange_declare

    async def publish(self, payload, exchange_name, routing_key, properties=None, mandatory=False, immediate=False):
        if isinstance(payload, str):
            warnings.warn("Str payload support will be removed in next release", DeprecationWarning)
            payload = payload.encode()

        if properties is None:
            properties = {}

        if self.publisher_confirms:
            delivery_tag = next(self.delivery_tag_iter)  # pylint: disable=stop-iteration-return
            fut = self._set_waiter('basic_server_ack_{}'.format(delivery_tag))

        method_request = pamqp.specification.Basic.Publish(
            exchange=exchange_name,
            routing_key=routing_key,
            mandatory=mandatory,
            immediate=immediate
        )
        await self._write_frame(self.channel_id, method_request, drain=False)

        properties = pamqp.specification.Basic.Properties(**properties)
        header_request = pamqp.header.ContentHeader(
            body_size=len(payload), properties=properties
        )
        await self._write_frame(self.channel_id, header_request, drain=False)

        # split the payload

        frame_max = self.protocol.server_frame_max or len(payload)
        for chunk in (payload[0+i:frame_max+i] for i in range(0, len(payload), frame_max)):
            content_request = pamqp.body.ContentBody(chunk)
            await self._write_frame(self.channel_id, content_request, drain=False)

        await self.protocol._drain()

        if self.publisher_confirms:
            await fut

    async def confirm_select(self, *, no_wait=False):
        if self.publisher_confirms:
            raise ValueError('publisher confirms already enabled')
        request = pamqp.specification.Confirm.Select(nowait=no_wait)

        return (await self._write_frame_awaiting_response(
            'confirm_select', self.channel_id, request, no_wait)
        )

    async def confirm_select_ok(self, frame):
        self.publisher_confirms = True
        self.delivery_tag_iter = count(1)
        fut = self._get_waiter('confirm_select')
        fut.set_result(True)
        logger.debug("Confirm selected")

    def add_cancellation_callback(self, callback):
        """Add a callback that is invoked when a consumer is cancelled.

        :param callback: function to call

        `callback` is called with the channel and consumer tag as positional
        parameters.  The callback can be either a plain callable or an
        asynchronous co-routine.

        """
        self.cancellation_callbacks.append(callback)



"""
    Tests the heartbeat methods
"""

import asyncio
import asynctest
from unittest import mock

from aioamqp.protocol import CLOSED

from . import testcase


class HeartbeatTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_heartbeat(self):
        with mock.patch.object(
                self.amqp, 'send_heartbeat', wraps=self.amqp.send_heartbeat
                ) as send_heartbeat:
            # reset both timers to 1) make them 'see' the new heartbeat value
            # 2) so that the mock is actually called back from the main loop
            self.amqp.server_heartbeat = 1
            self.amqp._heartbeat_timer_send_reset()
            self.amqp._heartbeat_timer_recv_reset()

            await asyncio.sleep(1.001)
            send_heartbeat.assert_called_once_with()

            await asyncio.sleep(1.001)
            self.assertEqual(self.amqp.state, CLOSED)



"""
    Amqp basic class tests
"""

import asyncio
import asynctest

from . import testcase
from .. import exceptions
from .. import properties


class QosTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_basic_qos_default_args(self):
        result = await self.channel.basic_qos()
        self.assertTrue(result)

    async def test_basic_qos(self):
        result = await self.channel.basic_qos(
            prefetch_size=0,
            prefetch_count=100,
            connection_global=False)

        self.assertTrue(result)

    async def test_basic_qos_prefetch_size(self):
        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.basic_qos(
                prefetch_size=10,
                prefetch_count=100,
                connection_global=False)

        self.assertEqual(cm.exception.code, 540)

    async def test_basic_qos_wrong_values(self):
        with self.assertRaises(TypeError):
            await self.channel.basic_qos(
                prefetch_size=100000,
                prefetch_count=1000000000,
                connection_global=False)


class BasicCancelTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_basic_cancel(self):

        async def callback(channel, body, envelope, _properties):
            pass

        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        await self.channel.queue_declare(queue_name)
        await self.channel.exchange_declare(exchange_name, type_name='direct')
        await self.channel.queue_bind(queue_name, exchange_name, routing_key='')
        result = await self.channel.basic_consume(callback, queue_name=queue_name)
        result = await self.channel.basic_cancel(result['consumer_tag'])

        result = await self.channel.publish("payload", exchange_name, routing_key='')

        await asyncio.sleep(5, loop=self.loop)

        result = await self.channel.queue_declare(queue_name, passive=True)
        self.assertEqual(result['message_count'], 1)
        self.assertEqual(result['consumer_count'], 0)


    async def test_basic_cancel_unknown_ctag(self):
        result = await self.channel.basic_cancel("unknown_ctag")
        self.assertTrue(result)


class BasicGetTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):


    async def test_basic_get(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.channel.queue_declare(queue_name)
        await self.channel.exchange_declare(exchange_name, type_name='direct')
        await self.channel.queue_bind(queue_name, exchange_name, routing_key=routing_key)

        await self.channel.publish("payload", exchange_name, routing_key=routing_key)

        result = await self.channel.basic_get(queue_name)
        self.assertEqual(result['routing_key'], routing_key)
        self.assertFalse(result['redelivered'])
        self.assertIn('delivery_tag', result)
        self.assertEqual(result['exchange_name'].split('.')[-1], exchange_name)
        self.assertEqual(result['message'], b'payload')
        self.assertIsInstance(result['properties'], properties.Properties)

    async def test_basic_get_empty(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''
        await self.channel.queue_declare(queue_name)
        await self.channel.exchange_declare(exchange_name, type_name='direct')
        await self.channel.queue_bind(queue_name, exchange_name, routing_key=routing_key)

        with self.assertRaises(exceptions.EmptyQueue):
            await self.channel.basic_get(queue_name)


class BasicDeliveryTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):


    async def publish(self, queue_name, exchange_name, routing_key, payload):
        await self.channel.queue_declare(queue_name, exclusive=False, no_wait=False)
        await self.channel.exchange_declare(exchange_name, type_name='fanout')
        await self.channel.queue_bind(queue_name, exchange_name, routing_key=routing_key)
        await self.channel.publish(payload, exchange_name, queue_name)



    async def test_ack_message(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(
            queue_name, exchange_name, routing_key, "payload"
        )

        qfuture = asyncio.Future(loop=self.loop)

        async def qcallback(channel, body, envelope, _properties):
            qfuture.set_result(envelope)

        await self.channel.basic_consume(qcallback, queue_name=queue_name)
        envelope = await qfuture

        await qfuture
        await self.channel.basic_client_ack(envelope.delivery_tag)

    async def test_basic_nack(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(
            queue_name, exchange_name, routing_key, "payload"
        )

        qfuture = asyncio.Future(loop=self.loop)

        async def qcallback(channel, body, envelope, _properties):
            await self.channel.basic_client_nack(
                envelope.delivery_tag, multiple=True, requeue=False
            )
            qfuture.set_result(True)

        await self.channel.basic_consume(qcallback, queue_name=queue_name)
        await qfuture

    async def test_basic_nack_norequeue(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(
            queue_name, exchange_name, routing_key, "payload"
        )

        qfuture = asyncio.Future(loop=self.loop)

        async def qcallback(channel, body, envelope, _properties):
            await self.channel.basic_client_nack(envelope.delivery_tag, requeue=False)
            qfuture.set_result(True)

        await self.channel.basic_consume(qcallback, queue_name=queue_name)
        await qfuture

    async def test_basic_nack_requeue(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''

        await self.publish(
            queue_name, exchange_name, routing_key, "payload"
        )

        qfuture = asyncio.Future(loop=self.loop)
        called = False

        async def qcallback(channel, body, envelope, _properties):
            nonlocal called
            if not called:
                called = True
                await self.channel.basic_client_nack(envelope.delivery_tag, requeue=True)
            else:
                await self.channel.basic_client_ack(envelope.delivery_tag)
                qfuture.set_result(True)

        await self.channel.basic_consume(qcallback, queue_name=queue_name)
        await qfuture


    async def test_basic_reject(self):
        queue_name = 'queue_name'
        exchange_name = 'exchange_name'
        routing_key = ''
        await self.publish(
            queue_name, exchange_name, routing_key, "payload"
        )

        qfuture = asyncio.Future(loop=self.loop)

        async def qcallback(channel, body, envelope, _properties):
            qfuture.set_result(envelope)

        await self.channel.basic_consume(qcallback, queue_name=queue_name)
        envelope = await qfuture

        await self.channel.basic_reject(envelope.delivery_tag)



"""
    Hello world `receive.py` example implementation using aioamqp.
    See the documentation for more informations.
"""

import asyncio
import aioamqp


async def callback(channel, body, envelope, properties):
    print(" [x] Received %r" % body)

async def receive():
    transport, protocol = await aioamqp.connect()
    channel = await protocol.channel()

    await channel.queue_declare(queue_name='hello')

    await channel.basic_consume(callback, queue_name='hello')


event_loop = asyncio.get_event_loop()
event_loop.run_until_complete(receive())
event_loop.run_forever()



"""Aioamqp tests utilities

Provides the test case to simplify testing
"""

import asyncio
import inspect
import logging
import os
import time
import uuid

import pyrabbit2.api

from .. import connect as aioamqp_connect
from .. import exceptions
from ..channel import Channel
from ..protocol import AmqpProtocol, OPEN


logger = logging.getLogger(__name__)


def use_full_name(f, arg_names):
    sig = inspect.signature(f)
    for arg_name in arg_names:
        if arg_name not in sig.parameters:
            raise ValueError('%s is not a valid argument name for function %s' % (arg_name, f.__qualname__))

    def wrapper(self, *args, **kw):
        ba = sig.bind_partial(self, *args, **kw)
        for param in sig.parameters.values():
            if param.name in arg_names and param.name in ba.arguments:
                ba.arguments[param.name] = self.full_name(ba.arguments[param.name])
        return f(*(ba.args), **(ba.kwargs))

    return wrapper


class ProxyChannel(Channel):
    def __init__(self, test_case, *args, **kw):
        super().__init__(*args, **kw)
        self.test_case = test_case
        self.test_case.register_channel(self)

    exchange_declare = use_full_name(Channel.exchange_declare, ['exchange_name'])
    exchange_delete = use_full_name(Channel.exchange_delete, ['exchange_name'])
    queue_declare = use_full_name(Channel.queue_declare, ['queue_name'])
    queue_delete = use_full_name(Channel.queue_delete, ['queue_name'])
    queue_bind = use_full_name(Channel.queue_bind, ['queue_name', 'exchange_name'])
    queue_unbind = use_full_name(Channel.queue_unbind, ['queue_name', 'exchange_name'])
    queue_purge = use_full_name(Channel.queue_purge, ['queue_name'])

    exchange_bind = use_full_name(Channel.exchange_bind, ['exchange_source', 'exchange_destination'])
    exchange_unbind = use_full_name(Channel.exchange_unbind, ['exchange_source', 'exchange_destination'])
    publish = use_full_name(Channel.publish, ['exchange_name'])
    basic_get = use_full_name(Channel.basic_get, ['queue_name'])
    basic_consume = use_full_name(Channel.basic_consume, ['queue_name'])

    def full_name(self, name):
        return self.test_case.full_name(name)


class ProxyAmqpProtocol(AmqpProtocol):
    def __init__(self, test_case, *args, **kw):
        super().__init__(*args, **kw)
        self.test_case = test_case

    def channel_factory(self, protocol, channel_id, return_callback=None):
        return ProxyChannel(self.test_case, protocol, channel_id,
                            return_callback=return_callback)
    CHANNEL_FACTORY = channel_factory


class RabbitTestCaseMixin:
    """TestCase with a rabbit running in background"""

    RABBIT_TIMEOUT = 1.0
    VHOST = 'test-aioamqp'

    def setUp(self):
        super().setUp()
        self.host = os.environ.get('AMQP_HOST', 'localhost')
        self.port = os.environ.get('AMQP_PORT', 5672)
        self.vhost = os.environ.get('AMQP_VHOST', self.VHOST + str(uuid.uuid4()))
        self.http_client = pyrabbit2.api.Client(
            '{HOST}:15672'.format(HOST=self.host), 'guest', 'guest', timeout=None,
        )

        self.amqps = []
        self.channels = []
        self.exchanges = {}
        self.queues = {}
        self.transports = []

        self.reset_vhost()

    def reset_vhost(self):
        try:
            self.http_client.delete_vhost(self.vhost)
        except Exception:  # pylint: disable=broad-except
            pass

        self.http_client.create_vhost(self.vhost)
        self.http_client.set_vhost_permissions(
            vname=self.vhost, username='guest', config='.*', rd='.*', wr='.*',
        )

        async def go():
            _transport, protocol = await self.create_amqp()
            channel = await self.create_channel(amqp=protocol)
            self.channels.append(channel)
        self.loop.run_until_complete(go())

    def tearDown(self):
        async def go():
            for queue_name, channel in self.queues.values():
                logger.debug('Delete queue %s', self.full_name(queue_name))
                await self.safe_queue_delete(queue_name, channel)
            for exchange_name, channel in self.exchanges.values():
                logger.debug('Delete exchange %s', self.full_name(exchange_name))
                await self.safe_exchange_delete(exchange_name, channel)
            for amqp in self.amqps:
                if amqp.state != OPEN:
                    continue
                logger.debug('Delete amqp %s', amqp)
                await amqp.close()
                del amqp
        self.loop.run_until_complete(go())

        try:
            self.http_client.delete_vhost(self.vhost)
        except Exception:  # pylint: disable=broad-except
            pass

        super().tearDown()

    @property
    def amqp(self):
        return self.amqps[0]

    @property
    def channel(self):
        return self.channels[0]

    def server_version(self, amqp=None):
        if amqp is None:
            amqp = self.amqp

        server_version = tuple(int(x) for x in amqp.server_properties['version'].decode().split('.'))
        return server_version

    async def check_exchange_exists(self, exchange_name):
        """Check if the exchange exist"""
        try:
            await self.exchange_declare(exchange_name, passive=True)
        except exceptions.ChannelClosed:
            return False

        return True

    async def assertExchangeExists(self, exchange_name):
        if not self.check_exchange_exists(exchange_name):
            self.fail("Exchange {} does not exists".format(exchange_name))

    async def check_queue_exists(self, queue_name):
        """Check if the queue exist"""
        try:
            await self.queue_declare(queue_name, passive=True)
        except exceptions.ChannelClosed:
            return False

        return True

    async def assertQueueExists(self, queue_name):
        if not self.check_queue_exists(queue_name):
            self.fail("Queue {} does not exists".format(queue_name))

    def list_queues(self, vhost=None, fully_qualified_name=False):
        # wait for the http client to get the correct state of the queue
        time.sleep(int(os.environ.get('AMQP_REFRESH_TIME', 6)))
        queues_list = self.http_client.get_queues(vhost=vhost or self.vhost)
        queues = {}
        for queue_info in queues_list:
            queue_name = queue_info['name']
            if fully_qualified_name is False:
                queue_name = self.local_name(queue_info['name'])
                queue_info['name'] = queue_name

            queues[queue_name] = queue_info
        return queues

    async def safe_queue_delete(self, queue_name, channel=None):
        """Delete the queue but does not raise any exception if it fails

        The operation has a timeout as well.
        """
        channel = channel or self.channel
        full_queue_name = self.full_name(queue_name)
        try:
            await channel.queue_delete(full_queue_name, no_wait=False)
        except asyncio.TimeoutError:
            logger.warning('Timeout on queue %s deletion', full_queue_name, exc_info=True)
        except Exception:  # pylint: disable=broad-except
            logger.exception('Unexpected error on queue %s deletion', full_queue_name)

    async def safe_exchange_delete(self, exchange_name, channel=None):
        """Delete the exchange but does not raise any exception if it fails

        The operation has a timeout as well.
        """
        channel = channel or self.channel
        full_exchange_name = self.full_name(exchange_name)
        try:
            await channel.exchange_delete(full_exchange_name, no_wait=False)
        except asyncio.TimeoutError:
            logger.warning('Timeout on exchange %s deletion', full_exchange_name, exc_info=True)
        except Exception:  # pylint: disable=broad-except
            logger.exception('Unexpected error on exchange %s deletion', full_exchange_name)

    def full_name(self, name):
        if self.is_full_name(name):
            return name
        return self.id() + '.' + name

    def local_name(self, name):
        if self.is_full_name(name):
            return name[len(self.id()) + 1:]  # +1 because of the '.'
        return name

    def is_full_name(self, name):
        return name.startswith(self.id())

    async def queue_declare(self, queue_name, *args, channel=None, safe_delete_before=True, **kw):
        channel = channel or self.channel
        if safe_delete_before:
            await self.safe_queue_delete(queue_name, channel=channel)
        # prefix queue_name with the test name
        full_queue_name = self.full_name(queue_name)
        try:
            rep = await channel.queue_declare(full_queue_name, *args, **kw)

        finally:
            self.queues[queue_name] = (queue_name, channel)
        return rep

    async def exchange_declare(self, exchange_name, *args, channel=None, safe_delete_before=True, **kw):
        channel = channel or self.channel
        if safe_delete_before:
            await self.safe_exchange_delete(exchange_name, channel=channel)
        # prefix exchange name
        full_exchange_name = self.full_name(exchange_name)
        try:
            rep = await channel.exchange_declare(full_exchange_name, *args, **kw)
        finally:
            self.exchanges[exchange_name] = (exchange_name, channel)
        return rep

    def register_channel(self, channel):
        self.channels.append(channel)

    async def create_channel(self, amqp=None):
        amqp = amqp or self.amqp
        channel = await amqp.channel()
        return channel

    async def create_amqp(self, vhost=None):
        def protocol_factory(*args, **kw):
            return ProxyAmqpProtocol(self, *args, **kw)
        vhost = vhost or self.vhost
        transport, protocol = await aioamqp_connect(host=self.host, port=self.port, virtualhost=vhost,
            protocol_factory=protocol_factory, loop=self.loop)
        self.amqps.append(protocol)
        return transport, protocol



"""
    Amqp exchange class tests
"""

import asynctest

from . import testcase
from .. import exceptions


class ExchangeDeclareTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    _multiprocess_can_split_ = True

    async def test_exchange_direct_declare(self):
        result = await self.channel.exchange_declare(
            'exchange_name', type_name='direct')
        self.assertTrue(result)

    async def test_exchange_fanout_declare(self):
        result = await self.channel.exchange_declare(
            'exchange_name', type_name='fanout')
        self.assertTrue(result)

    async def test_exchange_topic_declare(self):
        result = await self.channel.exchange_declare(
            'exchange_name', type_name='topic')
        self.assertTrue(result)

    async def test_exchange_headers_declare(self):
        result = await self.channel.exchange_declare(
            'exchange_name', type_name='headers')
        self.assertTrue(result)

    async def test_exchange_declare_wrong_types(self):
        result = await self.channel.exchange_declare(
            'exchange_name', type_name='headers',
            auto_delete=True, durable=True)
        self.assertTrue(result)

        with self.assertRaises(exceptions.ChannelClosed):
            result = await self.channel.exchange_declare(
                'exchange_name', type_name='fanout',
                auto_delete=False, durable=False)

    async def test_exchange_declare_passive(self):
        result = await self.channel.exchange_declare(
            'exchange_name', type_name='headers',
            auto_delete=True, durable=True)
        self.assertTrue(result)
        result = await self.channel.exchange_declare(
            'exchange_name', type_name='headers',
            auto_delete=True, durable=True, passive=True)
        self.assertTrue(result)

        result = await self.channel.exchange_declare(
            'exchange_name', type_name='headers',
            auto_delete=False, durable=False, passive=True)
        self.assertTrue(result)


    async def test_exchange_declare_passive_does_not_exists(self):
        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.exchange_declare(
                'non_existant_exchange',
                type_name='headers',
                auto_delete=False, durable=False, passive=True)
        self.assertEqual(cm.exception.code, 404)

    async def test_exchange_declare_unknown_type(self):
        with self.assertRaises(exceptions.ChannelClosed):
            await self.channel.exchange_declare(
                'non_existant_exchange',
                type_name='unknown_type',
                auto_delete=False, durable=False, passive=True)


class ExchangeDelete(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_delete(self):
        exchange_name = 'exchange_name'
        await self.channel.exchange_declare(exchange_name, type_name='direct')
        result = await self.channel.exchange_delete(exchange_name)
        self.assertTrue(result)
        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.exchange_declare(
                exchange_name, type_name='direct', passive=True
            )

        self.assertEqual(cm.exception.code, 404)


    async def test_double_delete(self):
        exchange_name = 'exchange_name'
        await self.channel.exchange_declare(exchange_name, type_name='direct')
        result = await self.channel.exchange_delete(exchange_name)
        self.assertTrue(result)
        if self.server_version() < (3, 3, 5):
            with self.assertRaises(exceptions.ChannelClosed) as cm:
                await self.channel.exchange_delete(exchange_name)

            self.assertEqual(cm.exception.code, 404)

        else:
            # weird result from rabbitmq 3.3.5
            result = await self.channel.exchange_delete(exchange_name)
            self.assertTrue(result)

class ExchangeBind(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    async def test_exchange_bind(self):
        await self.channel.exchange_declare('exchange_destination', type_name='direct')
        await self.channel.exchange_declare('exchange_source', type_name='direct')

        result = await self.channel.exchange_bind(
            'exchange_destination', 'exchange_source', routing_key='')

        self.assertTrue(result)

    async def test_inexistant_exchange_bind(self):
        with self.assertRaises(exceptions.ChannelClosed) as cm:
            await self.channel.exchange_bind(
                'exchange_destination', 'exchange_source', routing_key='')

        self.assertEqual(cm.exception.code, 404)


class ExchangeUnbind(testcase.RabbitTestCaseMixin, asynctest.TestCase):


    async def test_exchange_unbind(self):
        ex_source = 'exchange_source'
        ex_destination = 'exchange_destination'
        await self.channel.exchange_declare(ex_destination, type_name='direct')
        await self.channel.exchange_declare(ex_source, type_name='direct')

        await self.channel.exchange_bind(
            ex_destination, ex_source, routing_key='')

        await self.channel.exchange_unbind(
            ex_destination, ex_source, routing_key='')

    async def test_exchange_unbind_reversed(self):
        ex_source = 'exchange_source'
        ex_destination = 'exchange_destination'
        await self.channel.exchange_declare(ex_destination, type_name='direct')
        await self.channel.exchange_declare(ex_source, type_name='direct')

        await self.channel.exchange_bind(
            ex_destination, ex_source, routing_key='')

        if self.server_version() < (3, 3, 5):
            with self.assertRaises(exceptions.ChannelClosed) as cm:
                result = await self.channel.exchange_unbind(
                    ex_source, ex_destination, routing_key='')

            self.assertEqual(cm.exception.code, 404)

        else:
            # weird result from rabbitmq 3.3.5
            result = await self.channel.exchange_unbind(ex_source, ex_destination, routing_key='')
            self.assertTrue(result)



#!/usr/bin/env python

import asyncio
import aioamqp

import sys


async def new_task():
    try:
        transport, protocol = await aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return


    channel = await protocol.channel()

    await channel.queue('task_queue', durable=True)

    message = ' '.join(sys.argv[1:]) or "Hello World!"

    await channel.basic_publish(
        payload=message,
        exchange_name='',
        routing_key='task_queue',
        properties={
            'delivery_mode': 2,
        },
    )
    print(" [x] Sent %r" % message,)

    await protocol.close()
    transport.close()


asyncio.get_event_loop().run_until_complete(new_task())



import asyncio
import asynctest

from . import testcase
from .. import exceptions


class CloseTestCase(testcase.RabbitTestCaseMixin, asynctest.TestCase):

    def setUp(self):
        super().setUp()
        self.consume_future = asyncio.Future(loop=self.loop)

    async def callback(self, body, envelope, properties):
        self.consume_future.set_result((body, envelope, properties))

    async def get_callback_result(self):
        await self.consume_future
        result = self.consume_future.result()
        self.consume_future = asyncio.Future(loop=self.loop)
        return result

    async def test_close(self):
        channel = await self.create_channel()
        self.assertTrue(channel.is_open)
        await channel.close()
        self.assertFalse(channel.is_open)

    async def test_multiple_close(self):
        channel = await self.create_channel()
        await channel.close()
        self.assertFalse(channel.is_open)
        with self.assertRaises(exceptions.ChannelClosed):
            await channel.close()

    async def test_cannot_publish_after_close(self):
        channel = self.channel
        await channel.close()
        with self.assertRaises(exceptions.ChannelClosed):
            await self.channel.publish("coucou", "my_e", "")

    async def test_cannot_declare_queue_after_close(self):
        channel = self.channel
        await channel.close()
        with self.assertRaises(exceptions.ChannelClosed):
            await self.channel.queue_declare("qq")

    async def test_cannot_consume_after_close(self):
        channel = self.channel
        await self.channel.queue_declare("q")
        await channel.close()
        with self.assertRaises(exceptions.ChannelClosed):
            await channel.basic_consume(self.callback)



"""
    Hello world `send.py` example implementation using aioamqp.
    See the documentation for more informations.

    If there is no queue listening for the routing key, the message will
    get returned.

"""

import asyncio
import aioamqp


async def handle_return(channel, body, envelope, properties):
    print('Got a returned message with routing key: {}.\n'
          'Return code: {}\n'
          'Return message: {}\n'
          'exchange: {}'.format(envelope.routing_key, envelope.reply_code,
                                envelope.reply_text, envelope.exchange_name))


async def send():
    transport, protocol = await aioamqp.connect()
    channel = await protocol.channel(return_callback=handle_return)

    await channel.queue_declare(queue_name='hello')

    await channel.basic_publish(
        payload='Hello World!',
        exchange_name='',
        routing_key='helo',  # typo on purpose, will cause the return
        mandatory=True,
    )

    print(" [x] Sent 'Hello World!'")
    await protocol.close()
    transport.close()


asyncio.get_event_loop().run_until_complete(send())



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-three-python.html

"""

import asyncio
import aioamqp
import random



async def callback(channel, body, envelope, properties):
    print(" [x] %r" % body)


async def receive_log():
    try:
        transport, protocol = await aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return


    channel = await protocol.channel()
    exchange_name = 'logs'

    await channel.exchange(exchange_name=exchange_name, type_name='fanout')

    # let RabbitMQ generate a random queue name
    result = await channel.queue(queue_name='', exclusive=True)

    queue_name = result['queue']
    await channel.queue_bind(exchange_name=exchange_name, queue_name=queue_name, routing_key='')

    print(' [*] Waiting for logs. To exit press CTRL+C')

    await channel.basic_consume(callback, queue_name=queue_name, no_ack=True)

event_loop = asyncio.get_event_loop()
event_loop.run_until_complete(receive_log())
event_loop.run_forever()



#!/usr/bin/env python
"""
    Rabbitmq.com pub/sub example

    https://www.rabbitmq.com/tutorials/tutorial-four-python.html
"""

import asyncio
import aioamqp

import sys


async def exchange_routing():
    try:
        transport, protocol = await aioamqp.connect('localhost', 5672)
    except aioamqp.AmqpClosedConnection:
        print("closed connections")
        return

    channel = await protocol.channel()
    exchange_name = 'direct_logs'
    severity = sys.argv[1] if len(sys.argv) > 1 else 'info'
    message = ' '.join(sys.argv[2:]) or 'Hello World!'

    await channel.exchange(exchange_name, 'direct')

    await channel.publish(message, exchange_name=exchange_name, routing_key=severity)
    print(" [x] Sent %r" % (message,))

    await protocol.close()
    transport.close()


asyncio.get_event_loop().run_until_complete(exchange_routing())
